from tensorflow.python.ops import math_ops
from tensorflow.python.ops import array_ops
from tensorflow.python.ops import state_ops
from tensorflow.python.framework import ops
from tensorflow.keras.optimizers import Nadam, Adam
import tensorflow as tf
from tensorflow.python.keras.callbacks import Callback

class _Nadam(Nadam):
    def __init__(self,info=None, *args, **kwargs):
        super().__init__(*args, **kwargs)
        if info is None:
           info = dict()
        self.dataset_len = info.get('dataset_len',1)
        self.batch_size = info.get('batch_size',1)
        kwargs['decay'] = kwargs.get('schedule_decay', 0)


    def _prepare_local(self, var_device, var_dtype, apply_state):
        iter_transfer_actor = tf.constant(5 * (self.dataset_len // self.batch_size), dtype=self.iterations.dtype)
        local_iterations = tf.mod(self.iterations, iter_transfer_actor)
        lr_t = array_ops.identity(self._get_hyper('learning_rate', var_dtype))
        beta_1_t = array_ops.identity(self._get_hyper('beta_1', var_dtype))
        beta_2_t = array_ops.identity(self._get_hyper('beta_2', var_dtype))
        local_step = math_ops.cast(local_iterations + 1, var_dtype)
        next_step = math_ops.cast(local_iterations + 2, var_dtype)

        decay_base = math_ops.cast(0.96, var_dtype)

        m_t = beta_1_t * (1. - 0.5 * (
            math_ops.pow(decay_base, self._initial_decay * local_step)))
        m_t_1 = beta_1_t * (1. - 0.5 * (
            math_ops.pow(decay_base, self._initial_decay * next_step)))

        m_schedule_new = math_ops.cast(self._m_cache_read, var_dtype) * m_t
        if var_dtype is self._m_cache.dtype:
            m_schedule_new = array_ops.identity(state_ops.assign(
                self._m_cache, m_schedule_new, use_locking=self._use_locking))
        m_schedule_next = m_schedule_new * m_t_1

        apply_state[(var_device, var_dtype)] = dict(
            lr_t=lr_t,
            neg_lr_t=-lr_t,
            epsilon=ops.convert_to_tensor(self.epsilon, var_dtype),
            beta_1_t=beta_1_t,
            beta_2_t=beta_2_t,
            m_t=m_t,
            m_t_1=m_t_1,

            one_minus_beta_1_t=1 - beta_1_t,
            one_minus_beta_2_t=1 - beta_2_t,
            one_minus_m_t=1. - m_t,
            one_minus_m_schedule_new=1. - m_schedule_new,
            one_minus_m_schedule_next=1. - m_schedule_next,
            v_t_prime_denominator=1. - math_ops.pow(beta_2_t, local_step),
        )



class _Adam(Adam):

    def _prepare_local(self, var_device, var_dtype, apply_state):
        super(Adam, self)._prepare_local(var_device, var_dtype, apply_state)

        tmp_step = self.iterations % 1
        local_step = math_ops.cast(tmp_step + 1, var_dtype)
        beta_1_t = array_ops.identity(self._get_hyper('beta_1', var_dtype))
        beta_2_t = array_ops.identity(self._get_hyper('beta_2', var_dtype))
        beta_1_power = math_ops.pow(beta_1_t, local_step)
        beta_2_power = math_ops.pow(beta_2_t, local_step)
        lr = apply_state[(var_device, var_dtype)]['lr_t']
        apply_state[(var_device, var_dtype)].update(dict(
            lr=lr,
            epsilon=ops.convert_to_tensor(self.epsilon, var_dtype),
            beta_1_t=beta_1_t,
            beta_1_power=beta_1_power,
            one_minus_beta_1_t=1 - beta_1_t,
            beta_2_t=beta_2_t,
            beta_2_power=beta_2_power,
            one_minus_beta_2_t=1 - beta_2_t
        ))







#-*- coding: UTF-8 -*

import warnings, time, os
warnings.filterwarnings('ignore')
from deliverable_model import load
from deliverable_model.request import Request
from tokenizer_tools.tagset.offset.corpus import Corpus
from tokenizer_tools.tagset.offset.document import Document
from mtnlpmodel.server.inference.run_inference import MtModelInference_Deliverable


class MtModelEvaluation_Deliverable(MtModelInference_Deliverable):
    @classmethod
    def inference_process(cls, model, input_data, input_ids):
        output = []
        for data, id in zip(input_data, input_ids):
            req = []
            req.append(data)
            request = Request(req)
            response = model.inference(request)
            tmp_result = response['data'][0].sequence
            if not isinstance(tmp_result, Document):
                tmp_result = Document(tmp_result.text, tmp_result.span_set)
            tmp_result.label = response['cls'][0][0]
            tmp_result.id = id
            output.append(tmp_result)
        corpus_inference = Corpus(output)
        return corpus_inference


    @classmethod
    def whats_different(cls, corpus1, corpus2):
        corpus1_differ_from_corpus2 = Corpus(sorted(corpus1.difference(corpus2), key=lambda doc: doc.id))
        corpus2_differ_from_corpus1 = Corpus(sorted(corpus2.difference(corpus1), key=lambda doc: doc.id))
        return (corpus1_differ_from_corpus2, corpus2_differ_from_corpus1)


    @classmethod
    def calc_acc(self, differ_corpus_tuples, test_corpus_length):
        eps = 1e-9
        test, infer = differ_corpus_tuples
        cls_differ_count = 0
        ner_differ_count = 0
        for test_sample, infer_sample in zip(test, infer):
            if test_sample.label != infer_sample.label and \
               test_sample.id == infer_sample.id:
                cls_differ_count += 1

            if test_sample.span_set != infer_sample.span_set and \
               test_sample.id == infer_sample.id:
                ner_differ_count += 1
        scores = (1-cls_differ_count/(test_corpus_length+eps), 1-ner_differ_count/(test_corpus_length+eps))
        return scores


    def _evaluation(self, corpus_test):
        model = load(self.config['model_filepath'])
        text_list = [''.join(sample.text) for sample in corpus_test]
        id_list = [sample.id for sample in corpus_test]
        corpus_infer = MtModelEvaluation_Deliverable.inference_process(model, text_list, id_list)

        # symantic_corpus = corpus_test.intersection(corpus_infer)
        differ_corpus_tuples = MtModelEvaluation_Deliverable.whats_different(corpus_test, corpus_infer)

        scores = MtModelEvaluation_Deliverable.calc_acc(differ_corpus_tuples, len(corpus_test))

        return scores, differ_corpus_tuples


    @classmethod
    def get_differ_info(cls, sample_tuple):
        test = sample_tuple[0]
        infer = sample_tuple[1]
        id = test.id
        label_test = test.label
        label_infer = infer.label
        infer_text = infer.text.copy()
        test_text = test.text.copy()
        test_spanset = test.span_set
        infer_spanset = infer.span_set

        count = 0
        for span in test_spanset:
            test_text[span.start+count] = '<' + test_text[span.start+count]
            test_text[span.end-1+count] += '>'
            test_text.insert(span.end+count, '(' + span.entity + ')')
            count += 1
        count =0
        for span in infer_spanset:
            infer_text[span.start+count] = '<' + infer_text[span.start+count]
            infer_text[span.end-1+count] += '>'
            infer_text.insert(span.end+count, '(' + span.entity + ')')
            count += 1

        test_text = ''.join(test_text)
        infer_text = ''.join(infer_text)
        output_test_line = 'id: {id} | ground: {test_text} => label: {label_test}\n'.format(id=id,
                                                                                            test_text=test_text,
                                                                                            label_test=label_test)
        output_infer_line = 'id: {id} | infer:  {infer_text} => label: {label_infer}\n'.format(id=id,
                                                                                               infer_text=infer_text,
                                                                                               label_infer=label_infer)
        return output_test_line, output_infer_line


    def save_result(self, scores, differ_corpus_tuples):
        output_metrics = []
        output_text = []
        output_metrics.append('CLS_ACC:  ' + str(scores[0]))
        output_metrics.append('NER_ACC:  ' + str(scores[1]))
        output_metrics.append('=='*40)
        output_metrics.append('Differences of Evaluation list below:')
        output_metrics.append('=='*40+'\n')
        output_metrics = '\n'.join(output_metrics)

        for sample_tuple in zip(differ_corpus_tuples[0], differ_corpus_tuples[1]):
            differ_info = MtModelEvaluation_Deliverable.get_differ_info(sample_tuple)
            output_text.append(''.join(differ_info))
        output_text = '\n'.join(output_text)

        output = []
        output.extend(output_metrics)
        output.extend(output_text)
        output_path = os.path.join(self.config['output_filepath'], 'evaluation_out.txt')
        with open(output_path, 'wt', encoding='utf-8') as f:
            f.writelines(output)
        print('Result has been saved in the path:')
        print('{}'.format(output_path))


    def __call__(self):
        corpus_test = Corpus.read_from_file(os.path.join(self.config['data_filepath'], self.config['data_filename']))
        scores, differ_corpus_tuples = self._evaluation(corpus_test)
        self.save_result(scores, differ_corpus_tuples)
        print('Evaluation has been done.')



if __name__ == '__main__':
    print('*****The Evaluation program is running now, please wait.*****\n')
    start = time.time()
    config_filepath = './configure.json'
    MtModelEvaluation_Deliverable(config_filepath).__call__()
    end = time.time()
    print('==>Time cost is', end - start, 's')


import random
import numpy as np
import tensorflow as tf
from tokenizer_tools.tagset.offset.corpus import Corpus
from tokenizer_tools.tagset.converter.offset_to_biluo import offset_to_biluo
from seq2annotation.input import generate_tagset, Lookuper, index_table_from_file


def read_from_file(path):
    with open(path, 'rt', encoding='utf-8') as f:
        output = [line.strip() for line in f.readlines()]
    return output


def _read_configure(config_filepath, return_empty=False):
    import os
    import pprint
    from pconf import Pconf
    from ioflow.configure.get_configure_path_from_argv import get_configure_path_from_argv
    from ioflow.configure.read_configure import find_best_file_candidate, guess_configure_file_type

    # set return_empty to True for not read config from env
    # which can prevent unexpected result
    # e.g. './configure.json' is not for this app, but for other using
    if return_empty:
        return {}

    config_fileprefix = os.path.splitext(os.path.split(config_filepath)[-1])[0]

    default_configure_candidate = [
        ".".join([config_fileprefix, ext]) for ext in ["yaml", "yml", "json"]
    ]
    builtin_configure_candidate = [
        ".".join(["./builtin_configure", ext]) for ext in ["yaml", "yml", "json"]
    ]

    default_configure = find_best_file_candidate(default_configure_candidate)
    builtin_configure = find_best_file_candidate(builtin_configure_candidate)

    active_configure_file = get_configure_path_from_argv()
    if not active_configure_file:
        active_configure_file = os.getenv("_DEFAULT_CONFIG_FILE", default_configure)

    builtin_configure_file = os.getenv("_BUILTIN_CONFIG_FILE", builtin_configure)

    # Note: this is a safeguard, before using any Pconf function, do execute this
    # In case former Pconf usage influence current usage
    # which will lead to a hidden and wired bug
    Pconf.clear()

    # disable read configure from environment
    # Pconf.env()

    active_configure_file_abs_path = os.path.realpath(active_configure_file)

    if not os.path.exists(active_configure_file):
        msg = "default configure file is not found! CWD: {}; activate_config: {}; builtin_configure: {}".format(
            os.getcwd(), active_configure_file, builtin_configure_file
        )
        print(msg)
        raise ValueError(msg)
    else:
        print(
            ">>> Using configure read from file: {}".format(
                active_configure_file_abs_path
            )
        )

    file_encoding = guess_configure_file_type(active_configure_file_abs_path)
    Pconf.file(active_configure_file, file_encoding)

    # try loading builtin configure file
    if builtin_configure_file and os.path.exists(builtin_configure_file):
        print("loading builtin configure from {}".format(builtin_configure_file))
        file_encoding = guess_configure_file_type(builtin_configure_file)
        Pconf.file(builtin_configure_file, encoding=file_encoding)
    else:
        print(">>> builtin configure file is not found!")

    # Get all the config values parsed from the sources
    config = Pconf.get()

    # NOTE: clean Pconf for later brand new use
    Pconf.clear()

    print("++" * 8, "configure", "++" * 8)
    pprint.pprint(config)

    return config


def get_label_from_corpus(corpus):
    labels = set()
    for sample in corpus:
        labels.add(sample.label)
    return sorted(list(labels))


def get_tag_from_corpus(corpus):
    tags = set()
    for sample in corpus:
        for span in sample.span_set:
            tags.add(span.entity)
    return sorted(list(tags))


def build_vacablookuper_from_list(*lists):
    char_list = []
    for ls in lists:
        char_list.extend(ls)
    char_list = sorted(list(set(char_list)))
    if char_list[0]!='<pad>':
        char_list.insert(0, '<pad>')

    index_table = {}
    index_counter = 1
    for key in char_list:
        index_table[key] = index_counter
        index_counter += 1

    return Lookuper(index_table)


def build_vacablookuper_from_corpus(*corpus_tuples):
    char_list = []
    for corpus in corpus_tuples:
        for sample in corpus:
            char_list.extend(sample.text)
    char_list = sorted(list(set(char_list)))
    char_list.insert(0, '<pad>')

    index_table = {}
    index_counter = 0
    for key in char_list:
        index_table[key] = index_counter
        index_counter += 1

    return Lookuper(index_table)


def index_table_from_corpus(corpus=None):
    from seq2annotation.input import Lookuper
    index_table = {}
    tmp_text_list = [sample.text for sample in corpus]
    text_list = []
    for text in tmp_text_list:
        text_list.extend(text)
    text_list = sorted(set(text_list))
    for index, word in enumerate(text_list):
        index_table[word] = index+1
    return Lookuper(index_table)


def random_padding_to_samesize(ner_data_tuple, cls_data_tuple):
    ner_train_data, ner_eval_data = ner_data_tuple
    cls_train_data, cls_eval_data = cls_data_tuple
    if len(ner_train_data)>len(cls_train_data):
        padding_samples = random.sample(cls_train_data, (len(ner_train_data)-len(cls_train_data)))
        cls_train_data.extend(padding_samples)
    else:
        padding_samples = random.sample(ner_train_data, (len(cls_train_data) - len(ner_train_data)))
        ner_train_data.extend(padding_samples)

    if len(ner_eval_data)>len(cls_eval_data):
        padding_samples = random.sample(cls_eval_data, (len(ner_eval_data)-len(cls_eval_data)))
        cls_eval_data.extend(padding_samples)
    else:
        padding_samples = random.sample(ner_eval_data, (len(cls_eval_data) - len(ner_eval_data)))
        ner_eval_data.extend(padding_samples)

    ner_processed_tuple = (ner_train_data, ner_eval_data)
    cls_processed_tuple = (cls_train_data, cls_eval_data)
    return ner_processed_tuple, cls_processed_tuple


def random_sampling_to_samesize(ner_data_tuple, cls_data_tuple):
    ner_train_data, ner_eval_data = ner_data_tuple
    cls_train_data, cls_eval_data = cls_data_tuple

    if len(ner_train_data)>len(cls_train_data):
        ner_train_data = random.sample(ner_train_data, len(cls_train_data))
    else:
        cls_train_data = random.sample(cls_train_data, len(ner_train_data))

    if len(ner_eval_data)>len(cls_eval_data):
        ner_eval_data = random.sample(ner_eval_data, len(cls_eval_data))
    else:
        cls_eval_data = random.sample(cls_eval_data, len(ner_eval_data))

    ner_processed_tuple = (ner_train_data, ner_eval_data)
    cls_processed_tuple = (cls_train_data, cls_eval_data)
    return ner_processed_tuple, cls_processed_tuple


def input_data_process(config, **hyperparams):
    # read NER/CLS individually (only support *.conllx)
    input_mode = config['input_mode']
    if input_mode == 'multi':
        # multi input
        data_ner = Corpus.read_from_file(config['ner_data'])
        data_cls = Corpus.read_from_file(config['cls_data'])
    else:
        # single input
        data = Corpus.read_from_file(config['data'])
        data_ner = data
        data_cls = data

    # get train/test corpus
    ner_tags = get_tag_from_corpus(data_ner)
    cls_labels = get_label_from_corpus(data_cls)

    test_ratio = config['test_ratio']
    ner_train_data, ner_eval_data = data_ner.train_test_split(test_size=test_ratio, random_state=50)
    cls_train_data, cls_eval_data = data_cls.train_test_split(test_size=test_ratio, random_state=50)

    ner_data_tuple, cls_data_tuple = random_sampling_to_samesize((ner_train_data, ner_eval_data), # mainly for multi input
                                                                 (cls_train_data,cls_eval_data))  # make sure cls & ner have
    # same size dataset
    ner_train_data, ner_eval_data = ner_data_tuple
    cls_train_data, cls_eval_data = cls_data_tuple

    # build lookupers
    ner_tag_lookuper = Lookuper({v: i for i, v in enumerate(generate_tagset(ner_tags))})
    cls_label_lookuper = Lookuper({v: i for i, v in enumerate(cls_labels)})

    vocab_data_file = config.get("vocabulary_file", None)

    if not vocab_data_file:
        # get vacab_data for corpus
        vocabulary_lookuper = build_vacablookuper_from_corpus(*(data_ner, data_cls))  # from corpus
    else:
        vocabulary_lookuper = index_table_from_file(vocab_data_file)  # from vocab_file

    # ner (data&tag) str->int
    def ner_preprocss(data, maxlen, cls_info_len):
        raw_x_ner = []
        raw_y_ner = []

        for offset_data in data:
            tags = offset_to_biluo(offset_data)

            if config['decoder_style']=='BIO':
                for index, tag in enumerate(tags):
                    tags[index] = tag.replace('U', 'B')
                    tags[index] = tags[index].replace('L', 'I')

            words = offset_data.text

            tag_ids = [ner_tag_lookuper.lookup(i) for i in tags]
            word_ids = [vocabulary_lookuper.lookup(i) for i in words]

            raw_x_ner.append(word_ids)
            raw_y_ner.append(tag_ids)

        if maxlen is None:
            maxlen = max(len(s) for s in raw_x_ner)

        maxlen_mt = maxlen + cls_info_len
        print(">>> maxlen: {}".format(maxlen))

        x_ner = tf.keras.preprocessing.sequence.pad_sequences(
            raw_x_ner, maxlen, padding="post"
        )  # right padding

        y_ner = tf.keras.preprocessing.sequence.pad_sequences(
            raw_y_ner, maxlen, value=0, padding="post"
        )

        y_ner = tf.keras.preprocessing.sequence.pad_sequences(
            y_ner, maxlen_mt, value=0, padding="pre"
        )

        return x_ner, y_ner

    # cls (data&label) str->int
    def cls_preprocss(data, maxlen, **kwargs):
        raw_x_cls = []
        raw_y_cls = []

        for offset_data in data:
            label = offset_data.label
            words = offset_data.text

            label_id = cls_label_lookuper.lookup(label)
            word_ids = [vocabulary_lookuper.lookup(i) for i in words]

            raw_x_cls.append(word_ids)
            raw_y_cls.append(label_id)

        if maxlen is None:
            maxlen = max(len(s) for s in raw_x_cls)

        print(">>> maxlen: {}".format(maxlen))

        x_cls = tf.keras.preprocessing.sequence.pad_sequences(
            raw_x_cls, maxlen, padding="post"
        )  # right padding

        from tensorflow.keras.utils import to_categorical
        y_cls = np.array(raw_y_cls)
        y_cls = y_cls[:, np.newaxis]
        y_cls = to_categorical(y_cls, kwargs.get('cls_dims', 81))

        return x_cls, y_cls

    ner_train_x, ner_train_y = ner_preprocss(ner_train_data, hyperparams['MAX_SENTENCE_LEN'],
                                             hyperparams['CLS2NER_KEYWORD_LEN'])
    ner_test_x, ner_test_y = ner_preprocss(ner_eval_data, hyperparams['MAX_SENTENCE_LEN'],
                                           hyperparams['CLS2NER_KEYWORD_LEN'])

    cls_train_x, cls_train_y = cls_preprocss(cls_train_data, hyperparams['MAX_SENTENCE_LEN'],
                                             **{'cls_dims': cls_label_lookuper.size()})
    cls_test_x, cls_test_y = cls_preprocss(cls_eval_data, hyperparams['MAX_SENTENCE_LEN'],
                                           **{'cls_dims': cls_label_lookuper.size()})

    #cls_class_weight = get_class_weight(cls_train_data, cls_label_lookuper)

    output_dict = {'ner_train_x': ner_train_x, 'ner_train_y': ner_train_y,
                   'ner_test_x': ner_test_x, 'ner_test_y': ner_test_y,
                   'cls_train_x': cls_train_x, 'cls_train_y': cls_train_y,
                   'cls_test_x': cls_test_x, 'cls_test_y': cls_test_y,
                   'ner_tag_lookuper': ner_tag_lookuper,
                   'cls_label_lookuper': cls_label_lookuper,
                   'vocabulary_lookuper': vocabulary_lookuper,
                   }

    return output_dict


from collections import Counter
import numpy as np
import tensorflow as tf
from tensorflow.keras import backend as K


def class_counter(data, encoder):
    cls_counter = Counter([sample.label for sample in data])
    classes_num = [0]*encoder.size()
    for label in cls_counter.keys():
        index = encoder.lookup(label)
        classes_num[index] = cls_counter[label]
    return classes_num


def get_class_weight(data, encoder):
    cls_weight = dict()
    data_len = len(data)
    cls_counter = class_counter(data, encoder)
    for key, val in enumerate(cls_counter):
        cls_weight[key] = pow(data_len/(val+1), 0.5)

    return cls_weight



def focal_loss(gamma=2, alpha=0.25):
    '''for binary classification
       focal loss paper:
       https://arxiv.org/pdf/1708.02002.pdf
    '''
    def focal_loss_fixed(y_true, y_pred):
        eps = 1e-12
        y_pred = K.clip(y_pred, eps, 1. - eps)  # free from log0
        pt_1 = tf.where(tf.equal(y_true, 1), y_pred, tf.ones_like(y_pred))
        pt_0 = tf.where(tf.equal(y_true, 0), y_pred, tf.zeros_like(y_pred))
        return -K.sum(alpha * K.pow(1. - pt_1, gamma) * K.log(pt_1)) - K.sum(
            (1 - alpha) * K.pow(pt_0, gamma) * K.log(1. - pt_0))

    return focal_loss_fixed


def focal_loss_mt(classes_num, gamma=2):
    '''focal loss paper:
       https://arxiv.org/pdf/1708.02002.pdf
    '''
    def mt_focal_loss_fixed(y_true, y_pred):
        from tensorflow.python.ops import array_ops
        eps = 1e-9
        y_pred = K.clip(y_pred, eps, 1. - eps)  # free from log0

        zeros = array_ops.zeros_like(y_pred, dtype=y_pred.dtype)
        classes_weight = array_ops.zeros_like(y_pred, dtype=y_pred.dtype)
        total_num = float(sum(classes_num))
        classes_w_t1 = np.power([total_num / (ff + 1) for ff in classes_num], 1/2)
        classes_w_t1 = np.power([ ff/(max(classes_w_t1)+eps) for ff in classes_w_t1 ], 1/2)
        classes_w_tensor = tf.convert_to_tensor(classes_w_t1, dtype=tf.float32)
        classes_weight += classes_w_tensor

        alpha = array_ops.where(tf.greater(y_true, zeros), classes_weight, zeros)  # alpha is related with ratio of each class

        weight = tf.multiply(y_true, tf.pow(tf.subtract(1., y_pred), gamma))
        ce = tf.multiply(y_true, -tf.log(y_pred))
        fl = tf.multiply(alpha, tf.multiply(weight, ce))
        reduce_fl = tf.reduce_max(fl, axis=1)

        return tf.reduce_mean(reduce_fl)

    return mt_focal_loss_fixed


def _abandon_focal_loss_mt(classes_num, gamma=2): # there are some bugs in this function, so abandon
    '''focal loss paper:
       https://arxiv.org/pdf/1708.02002.pdf
    '''
    def mt_focal_loss_fixed(y_true, y_pred):
        from tensorflow.python.ops import array_ops
        eps = 1e-9
        y_pred = K.clip(y_pred, eps, 1. - eps)  # free from log0

        zeros = array_ops.zeros_like(y_pred, dtype=y_pred.dtype)
        classes_weight = array_ops.zeros_like(y_pred, dtype=y_pred.dtype)
        total_num = float(sum(classes_num))
        classes_w_t1 = np.power([total_num / (ff + 1) for ff in classes_num], 1/2)
        classes_w_t1 = np.power([ ff/(max(classes_w_t1)+eps) for ff in classes_w_t1 ], 1/2)
        classes_w_tensor = tf.convert_to_tensor(classes_w_t1, dtype=tf.float32)
        classes_weight += classes_w_tensor

        alpha = array_ops.where(tf.greater(y_true, zeros), classes_weight, zeros)  # alpha is related with ratio of each class

        pt_1 = tf.where(tf.equal(y_true, 1), y_pred, tf.ones_like(y_pred))
        pt_0 = tf.where(tf.equal(y_true, 0), y_pred, (tf.zeros_like(y_pred)+K.epsilon()))
        return -K.sum(alpha * K.pow(1. - pt_1, gamma) * K.log(pt_1)) - K.sum(
            (1 - alpha) * K.pow(pt_0, gamma) * K.log(1. - pt_0))

    return mt_focal_loss_fixed


def triplet_loss(y_true, y_pred, alpha=0.2):
    """
    Implementation of the triplet loss as defined by formula (3)

    Arguments:
    y_true -- true labels, required when you define a loss in Keras, you don't need it in this function.
    y_pred -- python list containing three objects:
            anchor -- the encodings for the anchor images, of shape (None, 128)
            positive -- the encodings for the positive images, of shape (None, 128)
            negative -- the encodings for the negative images, of shape (None, 128)

    Returns:
    loss -- real number, value of the loss
    """

    anchor, positive, negative = y_pred[0], y_pred[1], y_pred[2]

    # Step 1: Compute the (encoding) distance between the anchor and the positive, you will need to sum over axis=-1
    pos_dist = K.sum(K.square(anchor - positive), axis=-1)
    # Step 2: Compute the (encoding) distance between the anchor and the negative, you will need to sum over axis=-1
    neg_dist = K.sum(K.square(anchor - negative), axis=-1)
    # Step 3: subtract the two previous distances and add alpha.
    basic_loss = pos_dist - neg_dist + alpha
    # Step 4: Take the maximum of basic_loss and 0.0. Sum over the training examples.
    loss = K.sum(K.maximum(basic_loss, 0))

    return loss



import os
import pathlib
import shutil
from typing import Text, Any, Type
import random
from typing import Union, Callable, Dict

from deliverable_model.metacontent import MetaContent
from deliverable_model.builder import (
    DeliverableModelBuilder,
    MetadataBuilder,
    ProcessorBuilder,
    ModelBuilder,
)
from deliverable_model.builtin import LookupProcessor
from deliverable_model.builtin.processor import BILUOEncodeProcessor, PadProcessor
from deliverable_model.processor_base import ProcessorBase
from mtnlpmodel.utils.decoder_util import BIOEncodeProcessor

import numpy as np
from typing import Any
from deliverable_model.response import Response
from deliverable_model.converter_base import ConverterBase
from deliverable_model.request import Request


class RequestProcessorForVector(ProcessorBase):
    def preprocess(self, request: Request) -> Request:
        tmp_ner_query_list = []
        tmp_cls_query_list = []
        tmp_ner_query_list.append(request.query)
        tmp_cls_query_list.append(request.query)
        request['NER'] = tmp_ner_query_list
        request['CLS'] = tmp_cls_query_list

        return request

    @classmethod
    def load(cls, *args, **kwargs):
        return cls()


class RequestProcessor(ProcessorBase):
    def preprocess(self, request: Request) -> Request:
        request['NER'] = request.query
        request['CLS'] = request.query

        return request

    @classmethod
    def load(cls, *args, **kwargs):
        return cls()


class ConverterForRequest(ConverterBase):
    '''Now only support NER and CLS'''
    def call(self, request: Request) -> Any:
        request['input'] = request.query
        return request


class ConverterForMTRequest(ConverterBase):
    '''Now only support NER and CLS'''
    def call(self, request: Request) -> Any:
        request["CLS"] = request.query
        request["NER"] = request.query

        return [request["CLS"], request["NER"]]


class ConverterForVector(ConverterBase):
    def __call__(self, response) -> Response:
        return response



class ConverterForMTResponse(ConverterBase):
    def __call__(self, response) -> Response:
        response_data = response
        # ner branch
        ner_response = response_data[0].tolist()
        # cls branch
        cls_tmp_response = np.argmax(response_data[1])
        cls_tmp_restorer, cls_result_restorer = [], []
        cls_tmp_restorer.append(cls_tmp_response)
        cls_result_restorer.append(cls_tmp_restorer)
        result = Response([])
        result["cls"] = cls_result_restorer
        result.data = ner_response
        return result


class ConverterForMTResponse_VirtualPad(ConverterBase):
    def __init__(self, **kwargs):
        self.config = kwargs

    def __call__(self, response) -> Response:
        response_data = response
        # ner branch
        ner_response = response_data[0].tolist()
        ner_response[0][:self.config['prepad']] = []
        # cls branch
        cls_tmp_response = np.argmax(response_data[1])
        cls_tmp_restorer, cls_result_restorer = [], []
        cls_tmp_restorer.append(cls_tmp_response)
        cls_result_restorer.append(cls_tmp_restorer)

        result = Response([])
        result["cls"] = cls_result_restorer
        result.data = ner_response
        return result

    def get_config(self):
        return self.config


def mt_export_as_deliverable_model(
    output_dir,
    tensorflow_saved_model=None,
    converter_for_request: Union[None, Callable] = None,
    converter_for_response: Union[None, Callable] = None,
    keras_saved_model=None,
    keras_h5_model=None,
    meta_content_id="algorithmId-corpusId-configId-runId",
    lookup_tables: Dict = None,
    padding_parameter=None,
    addition_model_dependency=None,
    custom_object_dependency=None,
):
    # check parameters
    assert any(
        [tensorflow_saved_model, keras_saved_model, keras_h5_model]
    ), "one and only one of [tensorflow_saved_model, keras_saved_model, keras_h5_model] must be set up"
    assert (
        sum(
            int(bool(i))
            for i in [tensorflow_saved_model, keras_saved_model, keras_h5_model]
        )
        == 1
    ), "one and only one of [tensorflow_saved_model, keras_saved_model, keras_h5_model] must be set up"

    # default value
    addition_model_dependency = (
        [] if addition_model_dependency is None else addition_model_dependency
    )
    custom_object_dependency = (
        [] if custom_object_dependency is None else custom_object_dependency
    )

    # setup main object
    deliverable_model_builder = DeliverableModelBuilder(output_dir)

    # metadata builder
    metadata_builder = MetadataBuilder()

    meta_content = MetaContent(meta_content_id)

    metadata_builder.set_meta_content(meta_content)

    metadata_builder.save()

    # processor builder

    vocabulary_lookup_table = lookup_tables['vocab_lookup']
    tag_lookup_table = lookup_tables['tag_lookup']
    label_lookup_table = lookup_tables['label_lookup']

    processor_builder = ProcessorBuilder()

    decode_processor = BILUOEncodeProcessor()
    decoder_processor_handle = processor_builder.add_processor(decode_processor)

    pad_processor = PadProcessor(padding_parameter=padding_parameter)
    pad_processor_handle = processor_builder.add_processor(pad_processor)

    vocab_lookup_processor = LookupProcessor(vocabulary_lookup_table)
    vocab_lookup_processor_handle = processor_builder.add_processor(vocab_lookup_processor)

    tag_lookup_processor = LookupProcessor(tag_lookup_table)
    tag_lookup_processor_handle = processor_builder.add_processor(tag_lookup_processor)

    label_lookup_processor = LookupProcessor(label_lookup_table, **{"post_input_key": 'cls', "post_output_key": 'cls'})
    label_lookup_processor_handle = processor_builder.add_processor(label_lookup_processor)

    # # pre process: encoder[memory text] > lookup[str -> num] > pad[to fixed length]
    processor_builder.add_preprocess(decoder_processor_handle)
    processor_builder.add_preprocess(vocab_lookup_processor_handle)
    processor_builder.add_preprocess(pad_processor_handle)

    # # post process: lookup[num -> str] > encoder
    processor_builder.add_postprocess(tag_lookup_processor_handle)
    processor_builder.add_postprocess(label_lookup_processor_handle)
    processor_builder.add_postprocess(decoder_processor_handle)

    processor_builder.save()

    # model builder
    model_builder = ModelBuilder()
    model_builder.append_dependency(addition_model_dependency)
    model_builder.set_custom_object_dependency(custom_object_dependency)

    if converter_for_request:
        model_builder.add_converter_for_request(converter_for_request)

    if converter_for_response:
        model_builder.add_converter_for_response(converter_for_response)

    if tensorflow_saved_model:
        model_builder.add_tensorflow_saved_model(tensorflow_saved_model)
    elif keras_saved_model:
        model_builder.add_keras_saved_model(keras_saved_model)
    else:
        model_builder.add_keras_h5_model(keras_h5_model)

    model_builder.save()

    # compose all the parts
    deliverable_model_builder.add_processor(processor_builder)
    deliverable_model_builder.add_metadata(metadata_builder)
    deliverable_model_builder.add_model(model_builder)

    metadata = deliverable_model_builder.save()

    return metadata



def mtinput_export_as_deliverable_model(
    output_dir,
    tensorflow_saved_model=None,
    converter_for_request: Union[None, Callable] = None,
    converter_for_response: Union[None, Callable] = None,
    keras_saved_model=None,
    keras_h5_model=None,
    meta_content_id="algorithmId-corpusId-configId-runId",
    lookup_tables: Dict = None,
    padding_parameter=None,
    addition_model_dependency=None,
    custom_object_dependency=None,
    decoder_style='BILUO'
):
    # check parameters
    assert any(
        [tensorflow_saved_model, keras_saved_model, keras_h5_model]
    ), "one and only one of [tensorflow_saved_model, keras_saved_model, keras_h5_model] must be set up"
    assert (
        sum(
            int(bool(i))
            for i in [tensorflow_saved_model, keras_saved_model, keras_h5_model]
        )
        == 1
    ), "one and only one of [tensorflow_saved_model, keras_saved_model, keras_h5_model] must be set up"

    # default value
    addition_model_dependency = (
        [] if addition_model_dependency is None else addition_model_dependency
    )
    custom_object_dependency = (
        [] if custom_object_dependency is None else custom_object_dependency
    )

    # setup main object
    deliverable_model_builder = DeliverableModelBuilder(output_dir)

    # metadata builder
    metadata_builder = MetadataBuilder()

    meta_content = MetaContent(meta_content_id)

    metadata_builder.set_meta_content(meta_content)

    metadata_builder.save()

    # processor builder

    vocabulary_lookup_table = lookup_tables.get('vocab_lookup', None)
    tag_lookup_table = lookup_tables.get('tag_lookup', None)
    label_lookup_table = lookup_tables.get('label_lookup', None)

    processor_builder = ProcessorBuilder()

    if decoder_style=='BILUO':
        decode_processor = BILUOEncodeProcessor()
    else:
        decode_processor = BIOEncodeProcessor()

    decoder_processor_handle = processor_builder.add_processor(decode_processor)

    pad_processor = PadProcessor(padding_parameter=padding_parameter)
    pad_processor_handle = processor_builder.add_processor(pad_processor)

    request_processor = RequestProcessor()
    request_processor_handle = processor_builder.add_processor(request_processor)
    ner_vacab_lookup_processor = LookupProcessor(vocabulary_lookup_table)
    ner_vacab_lookup_processor.pre_input_key = 'NER'
    cls_vacab_lookup_processor = LookupProcessor(vocabulary_lookup_table)
    cls_vacab_lookup_processor.pre_input_key = 'CLS'
    ner_vocab_lookup_processor_handle = processor_builder.add_processor(ner_vacab_lookup_processor)
    cls_vocab_lookup_processor_handle = processor_builder.add_processor(cls_vacab_lookup_processor)

    tag_lookup_processor = LookupProcessor(tag_lookup_table)
    tag_lookup_processor_handle = processor_builder.add_processor(tag_lookup_processor)

    label_lookup_processor = LookupProcessor(label_lookup_table, **{"post_input_key": 'cls', "post_output_key": 'cls'})
    label_lookup_processor_handle = processor_builder.add_processor(label_lookup_processor)

    # # pre process: encoder[memory text] > lookup[str -> num] > pad[to fixed length]
    processor_builder.add_preprocess(request_processor_handle)
    processor_builder.add_preprocess(decoder_processor_handle)
    processor_builder.add_preprocess(ner_vocab_lookup_processor_handle)
    processor_builder.add_preprocess(cls_vocab_lookup_processor_handle)
    processor_builder.add_preprocess(pad_processor_handle)

    # # post process: lookup[num -> str] > encoder
    processor_builder.add_postprocess(tag_lookup_processor_handle)
    processor_builder.add_postprocess(label_lookup_processor_handle)
    processor_builder.add_postprocess(decoder_processor_handle)

    processor_builder.save()

    # model builder
    model_builder = ModelBuilder()
    model_builder.append_dependency(addition_model_dependency)
    model_builder.set_custom_object_dependency(custom_object_dependency)

    if converter_for_request:
        model_builder.add_converter_for_request(converter_for_request)

    if converter_for_response:
        model_builder.add_converter_for_response(converter_for_response)

    if tensorflow_saved_model:
        model_builder.add_tensorflow_saved_model(tensorflow_saved_model)
    elif keras_saved_model:
        model_builder.add_keras_saved_model(keras_saved_model)
    else:
        model_builder.add_keras_h5_model(keras_h5_model)

    model_builder.save()

    # compose all the parts
    deliverable_model_builder.add_processor(processor_builder)
    deliverable_model_builder.add_metadata(metadata_builder)
    deliverable_model_builder.add_model(model_builder)

    metadata = deliverable_model_builder.save()

    return metadata


def mtinput_export_as_vector_deliverable_model(
    output_dir,
    tensorflow_saved_model=None,
    converter_for_request: Union[None, Callable] = None,
    converter_for_response: Union[None, Callable] = None,
    keras_saved_model=None,
    keras_h5_model=None,
    meta_content_id="algorithmId-corpusId-configId-runId",
    lookup_tables: Dict = None,
    padding_parameter=None,
    addition_model_dependency=None,
    custom_object_dependency=None,
    decoder_style='BILUO'
):
    # check parameters
    assert any(
        [tensorflow_saved_model, keras_saved_model, keras_h5_model]
    ), "one and only one of [tensorflow_saved_model, keras_saved_model, keras_h5_model] must be set up"
    assert (
        sum(
            int(bool(i))
            for i in [tensorflow_saved_model, keras_saved_model, keras_h5_model]
        )
        == 1
    ), "one and only one of [tensorflow_saved_model, keras_saved_model, keras_h5_model] must be set up"

    # default value
    addition_model_dependency = (
        [] if addition_model_dependency is None else addition_model_dependency
    )
    custom_object_dependency = (
        [] if custom_object_dependency is None else custom_object_dependency
    )

    # setup main object
    deliverable_model_builder = DeliverableModelBuilder(output_dir)

    # metadata builder
    metadata_builder = MetadataBuilder()

    meta_content = MetaContent(meta_content_id)

    metadata_builder.set_meta_content(meta_content)

    metadata_builder.save()

    # processor builder

    vocabulary_lookup_table = lookup_tables.get('vocab_lookup', None)

    processor_builder = ProcessorBuilder()

    if decoder_style=='BILUO':
        decode_processor = BILUOEncodeProcessor()
    else:
        decode_processor = BIOEncodeProcessor()

    decoder_processor_handle = processor_builder.add_processor(decode_processor)

    pad_processor = PadProcessor(padding_parameter=padding_parameter)
    pad_processor_handle = processor_builder.add_processor(pad_processor)

    request_processor = RequestProcessorForVector()
    request_processor_handle = processor_builder.add_processor(request_processor)
    ner_vacab_lookup_processor = LookupProcessor(vocabulary_lookup_table)
    ner_vacab_lookup_processor.pre_input_key = 'NER'
    cls_vacab_lookup_processor = LookupProcessor(vocabulary_lookup_table)
    cls_vacab_lookup_processor.pre_input_key = 'CLS'
    ner_vocab_lookup_processor_handle = processor_builder.add_processor(ner_vacab_lookup_processor)
    cls_vocab_lookup_processor_handle = processor_builder.add_processor(cls_vacab_lookup_processor)

    # # pre process: encoder[memory text] > lookup[str -> num] > pad[to fixed length]
    processor_builder.add_preprocess(request_processor_handle)
    processor_builder.add_preprocess(decoder_processor_handle)
    processor_builder.add_preprocess(ner_vocab_lookup_processor_handle)
    processor_builder.add_preprocess(cls_vocab_lookup_processor_handle)
    processor_builder.add_preprocess(pad_processor_handle)

    # # # post process: lookup[num -> str] > encoder
    # processor_builder.add_postprocess(decoder_processor_handle)

    processor_builder.save()

    # model builder
    model_builder = ModelBuilder()
    model_builder.append_dependency(addition_model_dependency)
    model_builder.set_custom_object_dependency(custom_object_dependency)

    if converter_for_request:
        model_builder.add_converter_for_request(converter_for_request)

    if converter_for_response:
        model_builder.add_converter_for_response(converter_for_response)

    if tensorflow_saved_model:
        model_builder.add_tensorflow_saved_model(tensorflow_saved_model)
    elif keras_saved_model:
        model_builder.add_keras_saved_model(keras_saved_model)
    else:
        model_builder.add_keras_h5_model(keras_h5_model)

    model_builder.save()

    # compose all the parts
    deliverable_model_builder.add_processor(processor_builder)
    deliverable_model_builder.add_metadata(metadata_builder)
    deliverable_model_builder.add_model(model_builder)

    metadata = deliverable_model_builder.save()

    return metadata


def random_padding_to_samesize(ner_data_tuple, cls_data_tuple):
    ner_train_data, ner_eval_data = ner_data_tuple
    cls_train_data, cls_eval_data = cls_data_tuple
    if len(ner_train_data)>len(cls_train_data):
        padding_samples = random.sample(cls_train_data, (len(ner_train_data)-len(cls_train_data)))
        cls_train_data.extend(padding_samples)
    else:
        padding_samples = random.sample(ner_train_data, (len(cls_train_data) - len(ner_train_data)))
        ner_train_data.extend(padding_samples)

    if len(ner_eval_data)>len(cls_eval_data):
        padding_samples = random.sample(cls_eval_data, (len(ner_eval_data)-len(cls_eval_data)))
        cls_eval_data.extend(padding_samples)
    else:
        padding_samples = random.sample(ner_eval_data, (len(cls_eval_data) - len(ner_eval_data)))
        ner_eval_data.extend(padding_samples)

    ner_processed_tuple = (ner_train_data, ner_eval_data)
    cls_processed_tuple = (cls_train_data, cls_eval_data)
    return ner_processed_tuple, cls_processed_tuple


def random_sampling_to_samesize(ner_data_tuple, cls_data_tuple):
    ner_train_data, ner_eval_data = ner_data_tuple
    cls_train_data, cls_eval_data = cls_data_tuple

    if len(ner_train_data)>len(cls_train_data):
        ner_train_data = random.sample(ner_train_data, len(cls_train_data))
    else:
        cls_train_data = random.sample(cls_train_data, len(ner_train_data))

    if len(ner_eval_data)>len(cls_eval_data):
        ner_eval_data = random.sample(ner_eval_data, len(cls_eval_data))
    else:
        cls_eval_data = random.sample(cls_eval_data, len(ner_eval_data))

    ner_processed_tuple = (ner_train_data, ner_eval_data)
    cls_processed_tuple = (cls_train_data, cls_eval_data)
    return ner_processed_tuple, cls_processed_tuple


def remove_files_in_dir(data_dir):
    input_file_list = [i.absolute() for i in pathlib.Path(data_dir).iterdir() if i.is_file()]
    for i in input_file_list:
        os.remove(i)


def remove_content_in_dir(data_dir):
    input_file_list = pathlib.Path(data_dir).iterdir()
    for i in input_file_list:
        file_path = str(i.absolute())
        if i.is_dir():
            shutil.rmtree(file_path)
        else:
            os.remove(file_path)

def create_dir_if_needed(directory):
    # copied from https://stackoverflow.com/questions/273192/how-can-i-safely-create-a-nested-directory-in-python
    import shutil
    # if not os.path.exists(directory):
    if not os.path.exists(directory):
        # os.makedirs(directory)
        os.makedirs(directory)

    return directory


def create_or_rm_dir_if_needed(directory):
    # copied from https://stackoverflow.com/questions/273192/how-can-i-safely-create-a-nested-directory-in-python
    import shutil
    # if not os.path.exists(directory):
    if not os.path.exists(directory):
        # os.makedirs(directory)
        os.makedirs(directory)
    else:
        shutil.rmtree(directory)
        try:
            os.makedirs(directory)
        except:
            pass

    return directory


def create_file_dir_if_needed(file):
    directory = os.path.dirname(file)

    create_dir_if_needed(directory)

    return file


def join_path(a, b):
    return os.path.join(a, str(pathlib.PurePosixPath(b)))


def class_from_module_path(module_path: Text) -> Type[Any]:
    # copied from rasa_nlu (https://github.com/RasaHQ/rasa) @ rasa_nlu/utils/__init__.py
    """Given the module name and path of a class, tries to retrieve the class.

    The loaded class can be used to instantiate new objects. """
    import importlib

    # load the module, will raise ImportError if module cannot be loaded
    if "." in module_path:
        module_name, _, class_name = module_path.rpartition('.')
        m = importlib.import_module(module_name)
        # get the class, will raise AttributeError if class cannot be found
        return getattr(m, class_name)
    else:
        return globals()[module_path]


def load_hook(hook_config):
    hook_instances = []
    for i in hook_config:
        class_ = class_from_module_path(i['class'])
        hook_instances.append(class_(**i.get('params', {})))

    return hook_instances


from tensorflow.keras.utils import Sequence
class MakeSequence(Sequence):
    def __init__(self, data_sets, batch_size):
        self.ner_x, self.ner_y, self.cls_x, self.cls_y = data_sets
        self.batch_size = batch_size

    def __len__(self):
        return int(np.ceil(len(self.ner_x) / float(self.batch_size)))

    def __getitem__(self, idx):
        ner_x_batch_data = self.ner_x[(idx * self.batch_size):((idx + 1) * self.batch_size)]
        ner_y_batch_data = self.ner_y[(idx * self.batch_size):((idx + 1) * self.batch_size)]
        cls_x_batch_data = self.cls_x[(idx * self.batch_size):((idx + 1) * self.batch_size)]
        cls_y_batch_data = self.cls_y[(idx * self.batch_size):((idx + 1) * self.batch_size)]

        return ({'ner_input': ner_x_batch_data,
                'cls_input': cls_x_batch_data },
               {'crf': ner_y_batch_data,
                'CLS': cls_y_batch_data})



def to_categorical_strat_from_one(y, num_classes=None, dtype='float32'):
    """Converts a class vector (integers) to binary class matrix.

    E.g. for use with categorical_crossentropy.

    # Arguments
        y: class vector to be converted into a matrix
            (integers from 1 to num_classes).
        num_classes: total number of classes.
        dtype: The data type expected by the input, as a string
            (`float32`, `float64`, `int32`...)

    # Returns
        A binary matrix representation of the input. The classes axis
        is placed last.

    # Example

    ```python
    # Consider an array of 5 labels out of a set of 3 classes {1, 2, 3}:
    > labels
    array([1, 3, 1, 3, 1])
    # `to_categorical` converts this into a matrix with as many
    # columns as there are classes. The number of rows
    # stays the same.
    > to_categorical(labels)
    array([[ 1.,  0.,  0.],
           [ 0.,  0.,  1.],
           [ 0.,  1.,  0.],
           [ 0.,  0.,  1.],
           [ 1.,  0.,  0.]], dtype=float32)
    ```
    """

    y = np.array(y, dtype='int')
    input_shape = y.shape
    if input_shape and input_shape[-1] == 1 and len(input_shape) > 1:
        input_shape = tuple(input_shape[:-1])
    y = y.ravel()
    if not num_classes:
        num_classes = np.max(y) + 1
    n = y.shape[0]
    categorical = np.zeros((n, num_classes), dtype=dtype)
    categorical[np.arange(n), y-1] = 1
    output_shape = input_shape + (num_classes,)
    categorical = np.reshape(categorical, output_shape)
    return categorical


def load_ckpt(ckpt_path, model):
    if not os.path.exists(ckpt_path):
        return
    else:
        ckpts = os.listdir(ckpt_path)
        if not ckpts:
            return
        else:
            import tensorflow as tf
            latest_ckpt = tf.train.latest_checkpoint(ckpt_path)
            model.load_weights(latest_ckpt)
            print('Load ckpt from {}'.format(latest_ckpt))


# some models for classification
import math
import tensorflow as tf
from tf_crf_layer import keras_utils
from tensorflow.keras import backend as K
from tensorflow.keras.layers import Layer
from tensorflow.keras.layers import Embedding
from tensorflow.keras.activations import softplus, tanh
from tensorflow.python.keras import backend as K
from tensorflow.python.keras import constraints
from tensorflow.python.keras import initializers
from tensorflow.python.keras import regularizers
from tensorflow.python.ops import math_ops


def lstm_cls(input_layer, output_dim):
    from tensorflow.keras.layers import Dense, Dropout, Flatten
    cls_flat = Flatten()(input_layer)
    cls_dropout = Dropout(0.2)(cls_flat)
    output_layer = Dense(output_dim, activation='softmax', name='cls_Dense')(cls_dropout)

    return output_layer


def dilated_cnn_cls(input_layer, output_dim):
    from tensorflow.keras.layers import Conv1D, MaxPooling1D, BatchNormalization, Dense, Dropout, Flatten
    input_layer = BatchNormalization()(input_layer)
    cls_conv_emb = Conv1D(32, 3, activation='relu', padding='same')(input_layer)
    cls_conv_emb = Conv1D(64, 3, activation='relu', padding='same')(cls_conv_emb)
    cls_conv_emb = MaxPooling1D(2)(cls_conv_emb)
    cls_conv_emb = Conv1D(128, 3, activation='relu', dilation_rate=1, padding='same')(cls_conv_emb)
    cls_conv_emb = Conv1D(128, 3, activation='relu', dilation_rate=2, padding='same')(cls_conv_emb)
    cls_conv_emb = Conv1D(128, 3, activation='relu', dilation_rate=5, padding='same')(cls_conv_emb)
    cls_conv_emb = Conv1D(256, 1, activation='relu', padding='same')(cls_conv_emb)
    cls_conv_emb = MaxPooling1D(2)(cls_conv_emb)

    cls_flat = BatchNormalization()(cls_conv_emb)
    cls_flat = Flatten()(cls_flat)
    classification_dense = Dropout(0.2)(cls_flat)
    output_layer = Dense(output_dim, activation='softmax', name='CLS')(classification_dense)

    return output_layer


def textcnn_cls(input_layer, output_dim, outputlayer_name='CLS'):
    from tensorflow.keras.layers import Conv1D, MaxPooling1D, concatenate, Dense, Dropout, Flatten
    kernel_sizes = [2, 3, 4]
    pooling_out = []
    cls_conv_layer = input_layer
    for kernel_size in kernel_sizes:
        cls_conv_layer = Conv1D(filters=128, kernel_size=kernel_size, strides=1)(cls_conv_layer)
        cls_pooling_layer = MaxPooling1D(pool_size=int(cls_conv_layer.shape[1]))(cls_conv_layer)
        pooling_out.append(cls_pooling_layer)
        #print("kernel_size: %s \t c.shape: %s \t p.shape: %s" % (kernel_size, str(cls_conv_layer.shape), str(cls_pooling_layer.shape)))
    pool_output = concatenate([p for p in pooling_out])
    cls_flat = Flatten()(pool_output)
    cls_flat = Dropout(0.2)(cls_flat)
    cls_dense = Dense(output_dim, activation='softmax', name=outputlayer_name)(cls_flat)

    return cls_dense


def fasttext_cls(input_layer, output_dim):
    from tensorflow.keras.layers import Dense, Flatten, Conv1D
    # fast_text generally use embedding layer as its input_layer
    # directly connect dense to classify
    word_conv = Conv1D(64, 3, activation='relu', padding='same')(input_layer)
    word_conv = Conv1D(64, 3, activation='relu', padding='same')(word_conv)
    cls_flat = Flatten()(word_conv)
    return Dense(output_dim, activation='softmax')(cls_flat)


def get_paragraph_vector(input_layer):
    from tensorflow.keras.layers import Conv1D, Flatten ,MaxPooling1D
    embedding_layer = input_layer
    paragraph_vector = Conv1D(32, 3, activation='relu', padding='same')(embedding_layer)
    paragraph_vector = Conv1D(64, 3, activation='relu', padding='same')(paragraph_vector)
    paragraph_vector = MaxPooling1D(2)(paragraph_vector)
    paragraph_vector = Conv1D(128, 3, activation='relu', dilation_rate=1, padding='same')(paragraph_vector)
    paragraph_vector = Conv1D(128, 3, activation='relu', dilation_rate=2, padding='same')(paragraph_vector)
    paragraph_vector = Conv1D(128, 3, activation='relu', dilation_rate=5, padding='same')(paragraph_vector)
    paragraph_vector = Conv1D(256, 1, activation='relu', padding='same')(paragraph_vector)
    paragraph_vector = MaxPooling1D(2)(paragraph_vector)
    paragraph_vector = Flatten()(paragraph_vector)

    return paragraph_vector


def Discriminator(A_layer, onetask_output_shape, output_dtype='int32'):
    '''
        "A func to concate a task's output and another's input, in other words, tasks cascade."

    :param gen_output_shape:
            Define a layer's shape to generate a Input Layer to accept some layer's Output;
    :param A_layer:
            The layer you want to concate with, always be a Input Layer;
    :return:
            A Model you can use like a Layer, which concates a branch's output and another's input.
    '''
    from tensorflow.keras.layers import Input
    from tensorflow.keras.models import Model
    another_inputs = A_layer
    onetask_output = Input(onetask_output_shape, dtype=output_dtype)
    concatenated_input = tf.keras.layers.concatenate([onetask_output, another_inputs])
    # Now start applying Convolution Layers on concatenated_input
    # Deconvolution Layers
    return Model(inputs=[onetask_output, another_inputs], outputs=concatenated_input, name='Discriminator')


def Discriminator_new(onetask_output_shape, output_dtype='int32'):
    '''
        "A func to concate a task's output and another's input, in other words, tasks cascade."

    :param gen_output_shape:
            Define a layer's shape to generate a Input Layer to accept some layer's Output;
    :param A_layer:
            The layer you want to concate with, such as a Embedding Layer;
    :return:
            A Model you can use like a Layer, which concates a branch's output and another's input.
    '''
    from tensorflow.keras.layers import Input
    from tensorflow.keras.models import Model
    onetask_output = Input(onetask_output_shape, dtype=output_dtype)

    # Now start applying Convolution Layers on concatenated_input
    # Deconvolution Layers
    return Model(inputs=onetask_output, outputs=onetask_output, name='Discriminator')


def get_ner_cls_output_tensor_merge_embedding(mapping_size=5):
    ''' get ner branch's classification results, output is a Tensor.
    '''
    def get_tensor(ner_cls_layer):
        from tensorflow.python.ops import math_ops
        cls_index_tensor = math_ops.cast(tf.keras.backend.argmax(ner_cls_layer, axis=-1), 'int32')
        cls_index_tensor = tf.expand_dims(cls_index_tensor, -1)
        # print_op = tf.print(cls_index_tensor)
        # with tf.control_dependencies([print_op]):
        cls_index_tensor = tf.multiply(cls_index_tensor, tf.ones((1, mapping_size), dtype='int32'))
        return cls_index_tensor
    return get_tensor


def get_ner_cls_output_tensor_merge_input(mapping_size=5, **kwargs):
    def get_tensor(ner_cls_layer):
        import tensorflow as tf
        from tensorflow.python.ops import math_ops
        vocab_size = kwargs['vocab_size']
        label_size = kwargs['label_size']
        charlist_offset = tf.constant(vocab_size - label_size, dtype='int32')
        cls_index_tensor = tf.add(math_ops.cast(tf.keras.backend.argmax(ner_cls_layer, axis=-1), 'int32'), charlist_offset)
        cls_index_tensor = tf.expand_dims(cls_index_tensor, -1)
        cls_index_tensor = tf.multiply(cls_index_tensor, tf.ones((1, mapping_size), dtype='int32'))
        return cls_index_tensor
    return get_tensor


# Vitual Embedding
class VirtualEmbedding(Embedding):
    def __init__(self,
                 input_dim,
                 output_dim,
                 embeddings_initializer='uniform',
                 embeddings_regularizer=None,
                 activity_regularizer=None,
                 embeddings_constraint=None,
                 mask_zero=True,
                 input_length=None,
                 mask_length=None,
                 **kwargs):
        if 'input_shape' not in kwargs:
            if input_length:
                kwargs['input_shape'] = (input_length,)
            else:
                kwargs['input_shape'] = (None,)
        dtype = kwargs.pop('dtype', K.floatx())
        # We set autocast to False, as we do not want to cast floating- point inputs
        # to self.dtype. In call(), we cast to int32, and casting to self.dtype
        # before casting to int32 might cause the int32 values to be different due
        # to a loss of precision.
        kwargs['autocast'] = False
        super(Embedding, self).__init__(dtype=dtype, **kwargs)

        self.input_dim = input_dim
        self.output_dim = output_dim
        self.embeddings_initializer = initializers.get(embeddings_initializer)
        self.embeddings_regularizer = regularizers.get(embeddings_regularizer)
        self.activity_regularizer = regularizers.get(activity_regularizer)
        self.embeddings_constraint = constraints.get(embeddings_constraint)
        self.mask_zero = mask_zero
        self.supports_masking = mask_zero
        self.input_length = input_length
        self.mask_length = mask_length

    def compute_mask(self, inputs, mask=None):
        if not self.mask_zero:
            return None
        if not self.mask_length:
            return None
        if self.input_length>self.mask_length:
            mask_ones = tf.ones((tf.shape(inputs)[0], self.mask_length), dtype='int32')
            mask_zeros = tf.zeros((tf.shape(inputs)[0], self.input_length-self.mask_length), dtype='int32')
            tmp_mask = tf.concat([mask_ones, mask_zeros], axis=-1)
            return math_ops.not_equal(tmp_mask, 0)
        else:
            return math_ops.greater_equal(inputs, 0)
        # return self.keras_masks

    def get_config(self):
        config = {
            'input_dim': self.input_dim,
            'output_dim': self.output_dim,
            'embeddings_initializer':
                initializers.serialize(self.embeddings_initializer),
            'embeddings_regularizer':
                regularizers.serialize(self.embeddings_regularizer),
            'activity_regularizer':
                regularizers.serialize(self.activity_regularizer),
            'embeddings_constraint':
                constraints.serialize(self.embeddings_constraint),
            'mask_zero': self.mask_zero,
            'input_length': self.input_length,
            'mask_length': self.mask_length
        }
        base_config = super(Embedding, self).get_config()
        return dict(list(base_config.items()) + list(config.items()))


# ArcFace
# Original paper: https://arxiv.org/pdf/1801.07698.pdf
# Original implementation: https://github.com/deepinsight/insightface
# Adapted from tensorflow implementation: https://github.com/luckycallor/InsightFace-tensorflow

# keras version
@keras_utils.register_keras_custom_object
class ArcFace(Layer):
    '''Custom Keras layer implementing ArcFace including:
    1. Generation of embeddings
    2. Loss function
    3. Accuracy function
    cite: https://github.com/ktjonsson/keras-ArcFace/tree/master/src
    '''

    def __init__(self, class_num, margin=0.5, scale=64., **kwargs):
        self.class_num = int(class_num)
        self.margin = margin
        self.scale = scale
        assert self.scale > 0.0
        assert self.margin >= 0.0
        assert self.margin < (math.pi / 2)

        self.cos_m = tf.math.cos(margin)
        self.sin_m = tf.math.sin(margin)
        self.mm = self.sin_m * margin
        self.threshold = tf.math.cos(tf.constant(math.pi) - margin)
        super(ArcFace, self).__init__(**kwargs)

    def build(self, input_shape):
        # Create a trainable weight variable for this layer.
        self.kernel = self.add_weight(name='kernel',
                                      shape=(input_shape[1], self.class_num),
                                      initializer='glorot_normal',
                                      trainable=True)
        super(ArcFace, self).build(input_shape)  # Be sure to call this at the end

    def call(self, x):
        embeddings = tf.nn.l2_normalize(x, axis=1, name='normed_embeddings')
        weights = tf.nn.l2_normalize(self.kernel, axis=0, name='normed_weights')
        cos_t = tf.matmul(embeddings, weights, name='cos_t')
        logits = self.get_logits(cos_t)
        softmax_prob = tf.nn.softmax(logits, axis=-1)
        # print_op = tf.print(logits, softmax_prob)
        # with tf.control_dependencies([print_op]):
            #softmax_prob=tf.identity(softmax_prob)
        return softmax_prob

    def get_logits(self, y_pred):
        cos_t = y_pred
        cos_t2 = tf.square(cos_t, name='cos_2')
        sin_t2 = tf.subtract(1., cos_t2, name='sin_2')
        sin_t = tf.sqrt(sin_t2, name='sin_t')
        cos_mt = self.scale * tf.subtract(tf.multiply(cos_t, self.cos_m), tf.multiply(sin_t, self.sin_m), name='cos_mt')
        cond_v = cos_t - self.threshold
        cond = tf.cast(tf.nn.relu(cond_v, name='if_else'), dtype=tf.bool)
        keep_val = self.scale * (cos_t - self.mm)
        cos_mt_temp = tf.where(cond, cos_mt, keep_val)
        labels = K.argmax(cos_t2, axis=-1)
        mask = tf.one_hot(labels, depth=self.class_num, name='one_hot_mask')
        inv_mask = tf.subtract(1., mask, name='inverse_mask')
        s_cos_t = tf.multiply(self.scale, cos_t, name='scalar_cos_t')
        logits = tf.add(tf.multiply(s_cos_t, inv_mask), tf.multiply(cos_mt_temp, mask), name='arcface_logits')
        return logits

    def compute_output_shape(self, input_shape):
        return (input_shape[0], self.class_num)

    def get_config(self):
        config = {
            'class_num': self.class_num,
            'margin': self.margin,
            'scale': self.scale,
        }

        base_config = super().get_config().copy()
        return dict(list(base_config.items()) + list(config.items()))


@keras_utils.register_keras_custom_object
class Mish(Layer):
    def __init__(self, **kwargs):
        super(Mish, self).__init__(**kwargs)
        #self.supports_masking = True

    def call(self, inputs):
        return inputs * tanh(softplus(inputs))

    def get_config(self):
        config = super(Mish, self).get_config()
        return config

    def compute_output_shape(self, input_shape):
        '''
         compute_output_shape(self, input_shape)：为了能让Keras内部shape的匹配检查通过，
         这里需要重写compute_output_shape方法去覆盖父类中的同名方法，来保证输出shape是正确的。
         父类Layer中的compute_output_shape方法直接返回的是input_shape这明显是不对的，
         所以需要我们重写这个方法。所以这个方法也是4个要实现的基本方法之一。
        '''
        return input_shape





#-*- coding: UTF-8 -*

import warnings
warnings.filterwarnings('ignore')
import tqdm
import json, os, time
import numpy as np
import tensorflow as tf
from deliverable_model import load
from deliverable_model.request import Request
from mtnlpmodel.utils.model_util import ArcFace
from mtnlpmodel.utils.model_util import VirtualEmbedding
from tokenizer_tools.tagset.offset.corpus import Corpus
from mtnlpmodel.utils.model_util import Mish

tf.keras.utils.get_custom_objects()[Mish.__name__] = Mish
tf.keras.utils.get_custom_objects()[ArcFace.__name__] = ArcFace
tf.keras.utils.get_custom_objects()[VirtualEmbedding.__name__] = VirtualEmbedding

'''
    调用模型对输入数据进行推理，即采用模型对输入数据做实体识别标注，
    推理结果供人工复查。
'''

class CalcSemanticSimilarity:
    def __init__(self, config_filepath):
        self.config_filepath = config_filepath
        self.config = self.get_config()

        if not os.path.exists(self.config['output_filepath']):
            os.mkdir(self.config['output_filepath'])


    def get_config(self):
        '''
            从配置文件中读取配置信息
        '''
        with open(self.config_filepath, 'rb') as f:
            self.config = json.load(f)
        return self.config


    @staticmethod
    def generate_batch_input(input_data, batch_size):
        length = len(input_data)
        if length<batch_size:
            yield input_data
        else:
            for i in range(0, length, batch_size):
                yield(input_data[i: i+batch_size])


    @staticmethod
    def calc_euclidean_distance(v1, v2):
        return np.linalg.norm(v1-v2)/(np.linalg.norm(v1)+np.linalg.norm(v2))


    def save_results_to_file(self, output):
        results_filepath = os.path.join(self.config['output_filepath'], 'similarity_out.txt')
        with open(results_filepath, 'wt', encoding='utf-8') as f:
            f.write('\n'.join(output))


    def _inference(self, model, input_data):
        output = []
        batch_size = 1
        batches = CalcSemanticSimilarity.generate_batch_input(input_data, batch_size)
        for batch in tqdm.tqdm(batches):
            input_A = batch[0][0]
            input_B = batch[0][1]
            request_A = Request(input_A)
            request_B = Request(input_B)
            vector_A = model.inference(request_A)
            vector_B = model.inference(request_B)
            distance = CalcSemanticSimilarity.calc_euclidean_distance(vector_A, vector_B)
            similarity = 1-distance
            print_str = str(input_A)+' : '+str(input_B)+' = '+'{}'.format(round(similarity*100, 2))+'%'
            output.append(print_str)

        self.save_results_to_file(output)
        print('*** inference has been done, please check the result through the path below:')
        print('==> {}'.format(self.config['output_filepath']))
        return


    def get_datatuples_from_file(self):
        input_rawdata_filename = self.config['input_rawdata']
        input_file = os.path.join(self.config['data_filepath'], input_rawdata_filename)
        with open(input_file,'rt',encoding='utf-8') as f:
            input_data = [line.strip() for line in f.readlines()]
        data_tuple_list = []
        for data in input_data:
            data_tuple = tuple(data.split('|'))
            data_tuple_list.append(data_tuple)
        return data_tuple_list
            

    def calc_semantic_similarity(self):
        data_tuples = self.get_datatuples_from_file()
        model = load(self.config['model_filepath'])
        self._inference(model, data_tuples)


if __name__ == "__main__":
    print('*****The Similarity Calculation program is running now, please wait.*****\n')
    start = time.time()
    config_filepath = './configure.json'
    CalcSemanticSimilarity(config_filepath).calc_semantic_similarity()
    end = time.time()
    print('==> Time cost is', end - start, 's')


import os
import tensorflow as tf
from tensorflow.keras.backend import set_session
from tf_crf_layer.loss import ConditionalRandomFieldLoss
from tf_crf_layer.metrics import SequenceCorrectness
# tf.enable_eager_execution()
from mtnlpmodel.utils.input_process_util import _read_configure
from mtnlpmodel.utils.ctrldir_util import (create_dir_if_needed,
                                           create_file_dir_if_needed,
                                           create_or_rm_dir_if_needed)
from mtnlpmodel.utils.deliverablemodel_util import (ConverterForVector,
                                                    ConverterForMTRequest,
                                                    ConverterForMTResponse_VirtualPad,
                                                    mtinput_export_as_deliverable_model, 
                                                    mtinput_export_as_vector_deliverable_model)


def main():
    # get configure
    config = _read_configure("./configure.yaml")

    # get Parameters (controller)
    EPOCHS = config.get("epochs", 10)
    PRETRAIN_EPOCHS =config.get("pretrain_epochs", 5)
    BATCHSIZE = config.get("batch_size", 32)
    PRETRAIN_BATCHSIZE = config.get("pretrain_batchsize", 32)
    LEARNINGRATE = config.get("learning_rate", 0.001)
    MAX_SENTENCE_LEN = config.get("max_sentence_len", 25)
    LRDECAY = config.get('lr_decay', False)
    EARLYSTOP = config.get('early_stop', False)

    # get Parameters (model select)
    MODEL_CHOICE = config.get("model_choice", "VIRTUAL_EMBEDDING")

    # get Parameters (model structure)
    CLS2NER_KEYWORD_LEN = config.get("cls2ner_keyword_len", 5)
    EMBED_DIM = config.get("embedding_dim", 128)
    ARCLOSS = config.get("Arcloss", True)
    USE_ATTENTION_LAYER = config.get("use_attention_layer", False)
    BiLSTM_STACK_CONFIG = config.get("bilstm_stack_config", [])
    CRF_PARAMS = config.get("crf_params", {})
    DECODER_STYLE = config.get("decoder_style", "BILUO")


    # set gpu occupancy rate
    gpu_cfg = tf.ConfigProto()
    gpu_cfg.gpu_options.allocator_type = 'BFC'  # A "Best-fit with coalescing" algorithm, simplified from a version of dlmalloc.
    gpu_cfg.gpu_options.per_process_gpu_memory_fraction = 0.5
    gpu_cfg.gpu_options.allow_growth = True
    set_session(tf.Session(config=gpu_cfg))

    # get preprocessed input data dict
    from mtnlpmodel.utils.input_process_util import input_data_process
    # to build a fixed training environment, input data should be fixed.
    # input_data should be shuffled and remove duplication outside the trainer before running the program.
    # input data should be corpus(no duplication, shuffle well)

    if MODEL_CHOICE=='VIRTUAL_EMBEDDING' or MODEL_CHOICE=='CLS2NER_INPUT':  # different model structures have different input process way
        data_dict = input_data_process(config, **{'MAX_SENTENCE_LEN': MAX_SENTENCE_LEN,      # preprocess the input_data
                                       'CLS2NER_KEYWORD_LEN': CLS2NER_KEYWORD_LEN,})
    else:
        data_dict = input_data_process(config, **{'MAX_SENTENCE_LEN': MAX_SENTENCE_LEN,  # preprocess the input_data
                                                  'CLS2NER_KEYWORD_LEN': 0, })
        PRETRAIN_EPOCHS = 0

    # get lookupers
    ner_tag_lookuper = data_dict['ner_tag_lookuper']
    cls_label_lookuper = data_dict['cls_label_lookuper']
    vocabulary_lookuper = data_dict['vocabulary_lookuper']

    # get train/test data for training model
    ner_train_x, ner_train_y = data_dict['ner_train_x'], data_dict['ner_train_y']
    ner_test_x, ner_test_y = data_dict['ner_test_x'], data_dict['ner_test_y']

    cls_train_x, cls_train_y = data_dict['cls_train_x'], data_dict['cls_train_y']
    cls_test_x, cls_test_y = data_dict['cls_test_x'], data_dict['cls_test_y']


    # build model or finetuning
    from mtnlpmodel.core import build_model_multi_input
    params = {'EMBED_DIM': EMBED_DIM,
              'PRETRAIN_EPOCHS': PRETRAIN_EPOCHS,
              'BiLSTM_STACK_CONFIG': BiLSTM_STACK_CONFIG,
              'MAX_SENTENCE_LEN': MAX_SENTENCE_LEN,
              'CLS2NER_KEYWORD_LEN': CLS2NER_KEYWORD_LEN,
              'USE_ATTENTION_LAYER': USE_ATTENTION_LAYER,
              'Arcloss': ARCLOSS,
              'ner_tag_lookuper': ner_tag_lookuper,
              'cls_label_lookuper': cls_label_lookuper,
              'vocabulary_lookuper': vocabulary_lookuper,
              'CRF_PARAMS': CRF_PARAMS
             }
    model_choice = MODEL_CHOICE   # VIRTUAL_EMBEDDING, CLS2NER_INPUT, OTHER
    print("Model structure choosing {}".format(model_choice))

    # train the model by random initializer(make a fresh start to train a model)
    model, semantic_vector = build_model_multi_input(model_choice, **params)     # to build the model and get cls_vector
    
    model.summary()


    # build callbacks list
    callbacks_list = []

    tensorboard_callback = tf.keras.callbacks.TensorBoard(
        #log_dir=create_dir_if_needed(config["summary_log_dir"]),
        log_dir='.\\results\\summary_log_dir',  # for windows
        batch_size=BATCHSIZE,
    )
    callbacks_list.append(tensorboard_callback)

    checkpoint_callback = tf.keras.callbacks.ModelCheckpoint(
        os.path.join(create_dir_if_needed(config["model_dir"]), "cp-{epoch:04d}.ckpt"),
        load_weights_on_restart=True,
        verbose=1,
    )
    callbacks_list.append(checkpoint_callback)
    
    # early stop util
    if EARLYSTOP:
        early_stop = tf.keras.callbacks.EarlyStopping(monitor='val_loss',  # early stop index
                                                      patience=3,          # early stop delay epoch
                                                      verbose=2,           # display mode
                                                      mode='auto')
        callbacks_list.append(early_stop)

    #learning rate decay util
    if LRDECAY:
        reduce_lr = tf.keras.callbacks.ReduceLROnPlateau(monitor='val_loss', factor=0.75, patience=3, verbose=1,
                                                         mode='auto', epsilon=0.0001, cooldown=0, min_lr=0.00001)
        callbacks_list.append(reduce_lr)
    
    # ner_loss_func
    ner_loss_func = ConditionalRandomFieldLoss()
    
    # set optimizer
    adam_optimizer = tf.keras.optimizers.Adam(learning_rate=LEARNINGRATE, beta_1=0.9, beta_2=0.999, amsgrad=True)

    NER_out_name = 'crf'
    CLS_out_name = 'cls'

    # pretrain model -> train cls branch
    model.compile(optimizer=adam_optimizer,
                  loss={NER_out_name: ner_loss_func, CLS_out_name: 'categorical_crossentropy'},
                  loss_weights={NER_out_name: 0., CLS_out_name: 10.},  # set weight of loss
                  metrics={NER_out_name: SequenceCorrectness(), CLS_out_name: 'categorical_accuracy'})

    model.fit(
              {'ner_input': ner_train_x, 'cls_input': cls_train_x},
              {NER_out_name: ner_train_y, CLS_out_name: cls_train_y},
              epochs=PRETRAIN_EPOCHS,
              batch_size=PRETRAIN_BATCHSIZE,
              class_weight={NER_out_name: None, CLS_out_name: 'auto'},   # cls loss multiply the class weights
              validation_data=[{'ner_input': ner_test_x, 'cls_input': cls_test_x},
                               {NER_out_name: ner_test_y, CLS_out_name: cls_test_y}],
              callbacks=callbacks_list,)

    # train model
    model.compile(optimizer=adam_optimizer,
                  loss={NER_out_name: ner_loss_func, CLS_out_name: 'categorical_crossentropy'},
                  loss_weights={NER_out_name: 15., CLS_out_name: 10.},  # set weight of loss
                  metrics={NER_out_name: SequenceCorrectness(), CLS_out_name: 'categorical_accuracy'})
    
    model.fit(
            {'ner_input': ner_train_x, 'cls_input': cls_train_x},
            {NER_out_name: ner_train_y, CLS_out_name: cls_train_y},
            epochs=EPOCHS,
            batch_size=BATCHSIZE,
            class_weight={NER_out_name: None, CLS_out_name: 'auto'},   # cls loss multiply the class weights
            validation_data=[{'ner_input': ner_test_x, 'cls_input': cls_test_x},
                             {NER_out_name: ner_test_y, CLS_out_name: cls_test_y}],
            callbacks=callbacks_list,)


    # save model
    model.save(create_file_dir_if_needed(config["h5_model_file"]))

    model.save_weights(create_file_dir_if_needed(config["h5_weights_file"]))

    semantic_vector.save(create_file_dir_if_needed(config["vector_model_file"]))
    
    semantic_vector.save_weights(create_file_dir_if_needed(config["vector_weights_file"]))

    tf.keras.experimental.export_saved_model(
            model, create_or_rm_dir_if_needed(config["saved_model_dir"])
        )

    tf.keras.experimental.export_saved_model(
            semantic_vector, create_or_rm_dir_if_needed(config["vector_model_dir"])
        )

    # save cls&NER deliverable model
    mtinput_export_as_deliverable_model(
        create_dir_if_needed(config["deliverable_model_dir"]),
        keras_saved_model=config["saved_model_dir"],
        converter_for_request=ConverterForMTRequest(),
        converter_for_response=ConverterForMTResponse_VirtualPad(prepad=CLS2NER_KEYWORD_LEN),
        lookup_tables={'vocab_lookup':vocabulary_lookuper,
                       'tag_lookup':ner_tag_lookuper,
                       'label_lookup':cls_label_lookuper},
        padding_parameter={"maxlen": MAX_SENTENCE_LEN, "value": 0, "padding": "post"},
        addition_model_dependency=["tf-crf-layer"],
        custom_object_dependency=["tf_crf_layer"],
        decoder_style=DECODER_STYLE,
    )

    # save semantic vector deliverable model
    mtinput_export_as_vector_deliverable_model(
        create_dir_if_needed(config["vector_deliverable_model_dir"]),
        keras_saved_model=config["vector_model_dir"],
        converter_for_request=ConverterForMTRequest(),
        converter_for_response=ConverterForVector(), 
        lookup_tables={'vocab_lookup':vocabulary_lookuper},
        padding_parameter={"maxlen": MAX_SENTENCE_LEN, "value": 0, "padding": "post"},
        addition_model_dependency=["tf-crf-layer"],
        custom_object_dependency=["tf_crf_layer"],
        decoder_style=DECODER_STYLE,
    )



if __name__ == "__main__":
    main()


import os
import tensorflow as tf
from tensorflow.keras import Input
from tensorflow.keras.models import Model
from tensorflow.keras.layers import (Embedding,
                                     Flatten,
                                     Dropout,
                                     Dense,
                                     Lambda,
                                     Bidirectional,
                                     LSTM,
                                     LayerNormalization,)
from tf_crf_layer.layer import CRF
from tf_attention_layer.layers.global_attentioin_layer import GlobalAttentionLayer
from mtnlpmodel.utils.model_util import Mish


def cls_branch_single_input(arcloss_param, output_dims, feature_extractor, emb_layer, outputlayer_name='cls'):
    if arcloss_param:  # for arc-softmax loss
        from mtnlpmodel.utils.model_util import ArcFace
        with tf.keras.backend.name_scope("CLS_branch"):
            # cls branch
            cls_feature_layer = feature_extractor(emb_layer)
            cls_flat_lstm = Flatten()(cls_feature_layer)
            cls_flat = Dropout(0.25)(cls_flat_lstm)
            cls_vec_layer = Dense(128, name='arc_vector', activation='linear')
            cls_vector = cls_vec_layer(cls_flat)
            cls_vector = Mish()(cls_vector)
            cls_layer = ArcFace(output_dims, margin=0.2, name=outputlayer_name)
            cls_arc = cls_layer(cls_vector)
            cls_output = cls_arc
            # ner cls branch
            ner_cls_layer = cls_arc

    else:   # for softmax loss
        with tf.keras.backend.name_scope("CLS_branch"):
            # cls branch
            cls_feature_layer = feature_extractor(emb_layer)
            cls_flat_lstm = Flatten()(cls_feature_layer)
            cls_flat = Dropout(0.25)(cls_flat_lstm)
            cls_dense = Dense(output_dims, activation='softmax', name=outputlayer_name)
            cls_output = cls_dense(cls_flat)
            cls_vector = cls_output
            # ner cls branch
            ner_cls_layer = cls_vector

    return ner_cls_layer, cls_output, cls_vector, cls_feature_layer


def cls_branch(arcloss_param, output_dims, feature_extractor, cls_emb_layer, ner_emb_layer=None, outputlayer_name='cls'):
    if arcloss_param:  # for arc-softmax loss
        from mtnlpmodel.utils.model_util import ArcFace
        with tf.keras.backend.name_scope("CLS_branch"):
            # cls branch
            cls_feature_layer = feature_extractor(cls_emb_layer)
            cls_flat_lstm = Flatten()(cls_feature_layer)
            cls_flat = Dropout(0.25)(cls_flat_lstm)
            cls_vec_layer = Dense(128, name='arc_vector', activation='linear')
            cls_vector = cls_vec_layer(cls_flat)
            cls_vector = Mish()(cls_vector)
            cls_layer = ArcFace(output_dims, margin=0.2, name=outputlayer_name)
            cls_arc = cls_layer(cls_vector)
            cls_output = cls_arc
            # ner cls branch
            if ner_emb_layer is not None:
                ner_cls_feature_layer = feature_extractor(ner_emb_layer)
                ner_cls_flat_lstm = Flatten()(ner_cls_feature_layer)
                ner_cls_vec_layer = cls_vec_layer(ner_cls_flat_lstm)
                ner_cls_layer = cls_layer(ner_cls_vec_layer)
            else:
                ner_cls_layer = None

    else:   # for softmax loss
        with tf.keras.backend.name_scope("CLS_branch"):
            # cls branch
            cls_feature_layer = feature_extractor(cls_emb_layer)
            cls_flat_lstm = Flatten()(cls_feature_layer)
            cls_flat = Dropout(0.25)(cls_flat_lstm)
            cls_dense = Dense(output_dims, activation='softmax', name=outputlayer_name)
            cls_output = cls_dense(cls_flat)
            cls_vector = cls_output
            # ner cls branch
            if ner_emb_layer is not None:
                ner_cls_feature_layer = feature_extractor(ner_emb_layer)
                ner_cls_flat_lstm = Flatten()(ner_cls_feature_layer)
                ner_cls_layer = cls_dense(ner_cls_flat_lstm)
            else:
                ner_cls_layer = None

    return ner_cls_layer, cls_output, cls_vector


def build_model_single_input(model_choice, **hyperparams):
    from mtnlpmodel.utils.model_util import (get_ner_cls_output_tensor_merge_embedding,
                                             get_ner_cls_output_tensor_merge_input)
    # get hyperparams
    EMBED_DIM = hyperparams['EMBED_DIM']
    CRF_PARAMS = hyperparams['CRF_PARAMS']
    BiLSTM_STACK_CONFIG = hyperparams['BiLSTM_STACK_CONFIG']
    CLS2NER_KEYWORD_LEN = hyperparams['CLS2NER_KEYWORD_LEN']
    USE_ATTENTION_LAYER = hyperparams['USE_ATTENTION_LAYER']
    tag_size = hyperparams['ner_tag_lookuper'].size()
    label_size = hyperparams['cls_label_lookuper'].size()
    vocab_size = hyperparams['vocabulary_lookuper'].size()

    # input layer
    input_length = hyperparams['MAX_SENTENCE_LEN']
    input_layer = Input(shape=(input_length,), dtype='int32', name='input')

    # encoder
    if model_choice == 'VIRTUAL_EMBEDDING':  # cls_out embedding merged to ner_input_embedding as virtual embedding
        from mtnlpmodel.utils.model_util import VirtualEmbedding, Discriminator_new
        with tf.keras.backend.name_scope("Encoder"):
            embedding_layer_vocab = Embedding(vocab_size,
                                              EMBED_DIM,
                                              mask_zero=True,
                                              input_length=input_length,
                                              name='embedding_vocab'
                                              )
            embedding_layer_virtual = VirtualEmbedding(label_size,
                                                       EMBED_DIM,
                                                       mask_zero=True,
                                                       input_length=CLS2NER_KEYWORD_LEN,
                                                       mask_length=CLS2NER_KEYWORD_LEN,
                                                       name='embedding_virtual',
                                                       )

            embedding = embedding_layer_vocab(input_layer)

            embedding = Dropout(0.15)(embedding)  # just like random erase

        with tf.keras.backend.name_scope("Feature_extractor"):
            for bilstm_config in BiLSTM_STACK_CONFIG:
                biLSTM = Bidirectional(LSTM(return_sequences=True, **bilstm_config, name='biLSTM'))
            bilstm_extrator = biLSTM

        # classification branch
        ner_cls_layer, cls_output, cls_vector, _ = cls_branch_single_input(hyperparams['Arcloss'],
                                                           label_size, bilstm_extrator,
                                                           embedding,
                                                           outputlayer_name='cls')
        ner_cls_output_shape = get_ner_cls_output_tensor_merge_embedding(CLS2NER_KEYWORD_LEN)(ner_cls_layer).shape
        ner_cls_output_layer = Lambda(get_ner_cls_output_tensor_merge_embedding(CLS2NER_KEYWORD_LEN),
                                      ner_cls_output_shape)(ner_cls_layer)

        # classification output will be used as a keyword adding to input of NER
        discriminator = Discriminator_new(onetask_output_shape=(CLS2NER_KEYWORD_LEN,),
                                          output_dtype='int32')
        ner_cls_input_layer = discriminator(ner_cls_output_layer)
        ner_virtual_embedding = embedding_layer_virtual(ner_cls_input_layer)
        ner_merged_embedding = tf.keras.layers.concatenate([ner_virtual_embedding, embedding], axis=1)
        ner_branch_embedding = ner_merged_embedding

        ner_feature_layer = None

    elif model_choice=='CLS2NER_INPUT':   # cls_out merged to ner_input as virtual keywords
        from mtnlpmodel.utils.model_util import Discriminator
        from mtnlpmodel.utils.input_process_util import build_vacablookuper_from_list
        vocabs = list(hyperparams['vocabulary_lookuper'].inverse_index_table.values())
        cls_labels = list(hyperparams['cls_label_lookuper'].inverse_index_table.values())
        vocabs.extend(cls_labels)
        vocabulary_lookuper = build_vacablookuper_from_list(*vocabs)

        vocab_size = vocabulary_lookuper.size()
        with tf.keras.backend.name_scope("Encoder"):
            embedding_layer = Embedding(vocab_size,
                                        EMBED_DIM,
                                        mask_zero=True,
                                        input_length=input_length,
                                        )
            embedding = embedding_layer(input_layer)

            embedding = Dropout(0.15)(embedding)  # just like random erase

        with tf.keras.backend.name_scope("Feature_extractor"):
            for bilstm_config in BiLSTM_STACK_CONFIG:
                biLSTM = Bidirectional(LSTM(return_sequences=True, **bilstm_config, name='biLSTM'))
            bilstm_extrator = biLSTM

        # classification branch
        ner_cls_layer, cls_output, cls_vector, _ = cls_branch_single_input(hyperparams['Arcloss'],
                                                                       label_size, bilstm_extrator, embedding,
                                                                       outputlayer_name='cls')
        ner_cls_output_shape = get_ner_cls_output_tensor_merge_input(CLS2NER_KEYWORD_LEN,
                                                                     **{"vocab_size":vocab_size,
                                                                        "label_size":label_size})(ner_cls_layer).shape

        ner_cls_output_layer = Lambda(get_ner_cls_output_tensor_merge_input(
                                         CLS2NER_KEYWORD_LEN,
                                         **{"vocab_size": vocab_size, "label_size": label_size}),
                                      ner_cls_output_shape)(ner_cls_layer)

        # classification output will be used as a keyword adding to input of NER
        discriminator = Discriminator(input_layer, onetask_output_shape=(CLS2NER_KEYWORD_LEN,),
                                      output_dtype='int32')
        merged_ner_input_layer = discriminator([ner_cls_output_layer, input_layer])
        ner_branch_embedding = embedding_layer(merged_ner_input_layer)

        ner_feature_layer = None

    else: # task independent
        with tf.keras.backend.name_scope("Encoder"):
            embedding_layer = Embedding(vocab_size,
                                        EMBED_DIM,
                                        mask_zero=True,
                                        input_length=input_length,
                                        )

            embedding = embedding_layer(input_layer)

            embedding = Dropout(0.15)(embedding)    # just like random erase

        with tf.keras.backend.name_scope("Feature_extractor"):
            for bilstm_config in BiLSTM_STACK_CONFIG:
                biLSTM = Bidirectional(LSTM(return_sequences=True, **bilstm_config, name='biLSTM'))
            bilstm_extrator = biLSTM

        # classification branch
        _, cls_output, cls_vector, cls_feature_layer = cls_branch_single_input(hyperparams['Arcloss'],
                                                            label_size, bilstm_extrator,
                                                            embedding, outputlayer_name='cls')
        ner_branch_embedding = embedding
        ner_feature_layer = cls_feature_layer

    # NER branch
    with tf.keras.backend.name_scope("NER_branch"):
        if ner_feature_layer is None:
            # print_op = tf.print(ner_virtual_embedding._keras_mask, ner_embedding._keras_mask)
            # with tf.control_dependencies([print_op]):
            embedding_layer = LayerNormalization()(ner_branch_embedding)
            biLSTM = bilstm_extrator(embedding_layer)
            biLSTM = LayerNormalization()(biLSTM)
            if USE_ATTENTION_LAYER:
                biLSTM = GlobalAttentionLayer()(biLSTM)
            ner_output = CRF(tag_size, name="crf", **CRF_PARAMS)(biLSTM)
        else:
            biLSTM = ner_feature_layer
            biLSTM = LayerNormalization()(biLSTM)
            if USE_ATTENTION_LAYER:
                biLSTM = GlobalAttentionLayer()(biLSTM)
            ner_output = CRF(tag_size, name="crf", **CRF_PARAMS)(biLSTM)

    # merge NER and Classification
    model = Model(inputs=[input_layer], outputs=[ner_output, cls_output])
    semantic_vector = Model(inputs=[input_layer], outputs=cls_vector)

    return model, semantic_vector


def build_model_multi_input(model_choice, **hyperparams):
    from mtnlpmodel.utils.model_util import (get_ner_cls_output_tensor_merge_embedding,
                                             get_ner_cls_output_tensor_merge_input)
    # get hyperparams
    EMBED_DIM = hyperparams['EMBED_DIM']
    CRF_PARAMS = hyperparams['CRF_PARAMS']
    BiLSTM_STACK_CONFIG = hyperparams['BiLSTM_STACK_CONFIG']
    CLS2NER_KEYWORD_LEN = hyperparams['CLS2NER_KEYWORD_LEN']
    USE_ATTENTION_LAYER = hyperparams['USE_ATTENTION_LAYER']
    tag_size = hyperparams['ner_tag_lookuper'].size()
    label_size = hyperparams['cls_label_lookuper'].size()
    vocab_size = hyperparams['vocabulary_lookuper'].size()

    # input layer
    input_length = hyperparams['MAX_SENTENCE_LEN']
    ner_input_layer = Input(shape=(input_length,), dtype='int32', name='ner_input')
    cls_input_layer = Input(shape=(input_length,), dtype='int32', name='cls_input')

    # encoder
    if model_choice=='VIRTUAL_EMBEDDING':   # cls_out embedding merged to ner_input_embedding as virtual embedding
        from mtnlpmodel.utils.model_util import VirtualEmbedding, Discriminator_new
        with tf.keras.backend.name_scope("Encoder"):
            embedding_layer_vocab = Embedding(vocab_size,
                                              EMBED_DIM,
                                              mask_zero=True,
                                              input_length=input_length,
                                              name='embedding_vocab'
                                              )
            embedding_layer_virtual = VirtualEmbedding(label_size,
                                                       EMBED_DIM,
                                                       mask_zero=True,
                                                       input_length=CLS2NER_KEYWORD_LEN,
                                                       mask_length=CLS2NER_KEYWORD_LEN,
                                                       name='embedding_virtual',
                                                       )

            ner_embedding = embedding_layer_vocab(ner_input_layer)
            cls_embedding = embedding_layer_vocab(cls_input_layer)

            ner_embedding = Dropout(0.15)(ner_embedding)    # just like random erase
            cls_embedding = Dropout(0.15)(cls_embedding)

        with tf.keras.backend.name_scope("Feature_extractor"):
            for bilstm_config in BiLSTM_STACK_CONFIG:
                biLSTM = Bidirectional(LSTM(return_sequences=True, **bilstm_config, name='biLSTM'))
            bilstm_extrator = biLSTM

        # classification branch
        ner_cls_layer, cls_output, cls_vector = cls_branch(hyperparams['Arcloss'],
                                                           label_size, bilstm_extrator,
                                                           cls_embedding, ner_embedding,
                                                           outputlayer_name='cls')
        ner_cls_output_shape = get_ner_cls_output_tensor_merge_embedding(CLS2NER_KEYWORD_LEN)(ner_cls_layer).shape
        ner_cls_output_layer = Lambda(get_ner_cls_output_tensor_merge_embedding(CLS2NER_KEYWORD_LEN), ner_cls_output_shape)(ner_cls_layer)

        # classification output will be used as a keyword adding to input of NER
        discriminator = Discriminator_new(onetask_output_shape=(CLS2NER_KEYWORD_LEN,),
                                          output_dtype='int32')
        ner_cls_input_layer = discriminator(ner_cls_output_layer)
        ner_virtual_embedding = embedding_layer_virtual(ner_cls_input_layer)
        ner_merged_embedding = tf.keras.layers.concatenate([ner_virtual_embedding, ner_embedding], axis=1)
        ner_branch_embedding = ner_merged_embedding

    elif model_choice=='CLS2NER_INPUT':   # cls_out merged to ner_input as virtual keywords
        from mtnlpmodel.utils.model_util import Discriminator
        from mtnlpmodel.utils.input_process_util import build_vacablookuper_from_list
        vocabs = list(hyperparams['vocabulary_lookuper'].inverse_index_table.values())
        cls_labels = list(hyperparams['cls_label_lookuper'].inverse_index_table.values())
        vocabs.extend(cls_labels)
        vocabulary_lookuper = build_vacablookuper_from_list(*vocabs)

        vocab_size = vocabulary_lookuper.size()
        with tf.keras.backend.name_scope("Encoder"):
            embedding_layer = Embedding(vocab_size,
                                        EMBED_DIM,
                                        mask_zero=True,
                                        input_length=input_length,
                                        )
            ner_embedding = embedding_layer(ner_input_layer)
            cls_embedding = embedding_layer(cls_input_layer)

            ner_embedding = Dropout(0.15)(ner_embedding)  # just like random erase
            cls_embedding = Dropout(0.15)(cls_embedding)

        with tf.keras.backend.name_scope("Feature_extractor"):
            for bilstm_config in BiLSTM_STACK_CONFIG:
                biLSTM = Bidirectional(LSTM(return_sequences=True, **bilstm_config, name='biLSTM'))
            bilstm_extrator = biLSTM

        # classification branch
        ner_cls_layer, cls_output, cls_vector = cls_branch(hyperparams['Arcloss'],
                                                           label_size, bilstm_extrator,
                                                           cls_embedding, ner_embedding,
                                                           outputlayer_name='cls')
        ner_cls_output_shape = get_ner_cls_output_tensor_merge_input(CLS2NER_KEYWORD_LEN,
                                                                     **{"vocab_size":vocab_size,
                                                                        "label_size":label_size})(ner_cls_layer).shape

        ner_cls_output_layer = Lambda(get_ner_cls_output_tensor_merge_input(
                                         CLS2NER_KEYWORD_LEN,
                                         **{"vocab_size": vocab_size, "label_size": label_size}),
                                      ner_cls_output_shape)(ner_cls_layer)

        # classification output will be used as a keyword adding to input of NER
        discriminator = Discriminator(ner_input_layer, onetask_output_shape=(CLS2NER_KEYWORD_LEN,),
                                      output_dtype='int32')
        merged_ner_input_layer = discriminator([ner_cls_output_layer, ner_input_layer])
        ner_branch_embedding = embedding_layer(merged_ner_input_layer)

    else: # task independent
        with tf.keras.backend.name_scope("Encoder"):
            embedding_layer = Embedding(vocab_size,
                                        EMBED_DIM,
                                        mask_zero=True,
                                        input_length=input_length,
                                        )

            ner_embedding = embedding_layer(ner_input_layer)
            cls_embedding = embedding_layer(cls_input_layer)

            ner_embedding = Dropout(0.15)(ner_embedding)    # just like random erase
            cls_embedding = Dropout(0.15)(cls_embedding)

        with tf.keras.backend.name_scope("Feature_extractor"):
            for bilstm_config in BiLSTM_STACK_CONFIG:
                biLSTM = Bidirectional(LSTM(return_sequences=True, **bilstm_config, name='biLSTM'))
            bilstm_extrator = biLSTM

        # classification branch
        _, cls_output, cls_vector = cls_branch(hyperparams['Arcloss'],
                                               label_size, bilstm_extrator,
                                               cls_embedding, outputlayer_name='cls')
        ner_branch_embedding = ner_embedding

    # NER branch
    with tf.keras.backend.name_scope("NER_branch"):
        # print_op = tf.print(ner_virtual_embedding._keras_mask, ner_embedding._keras_mask)
        # with tf.control_dependencies([print_op]):
        embedding_layer = LayerNormalization()(ner_branch_embedding)
        biLSTM = bilstm_extrator(embedding_layer)
        biLSTM = LayerNormalization()(biLSTM)
        if USE_ATTENTION_LAYER:
            biLSTM = GlobalAttentionLayer()(biLSTM)
        ner_output = CRF(tag_size, name="crf", **CRF_PARAMS)(biLSTM)

    # merge NER and Classification
    model = Model(inputs=[ner_input_layer, cls_input_layer], outputs=[ner_output, cls_output])
    semantic_vector = Model(inputs=[ner_input_layer, cls_input_layer], outputs=cls_vector)

    return model, semantic_vector


def finetune_model(model_choice, model_weights_path, freeze_list, **hyperparams):
    from mtnlpmodel.utils.model_util import (get_ner_cls_output_tensor_merge_embedding,
                                             get_ner_cls_output_tensor_merge_input)
    # get hyperparams
    EMBED_DIM = hyperparams['EMBED_DIM']
    CRF_PARAMS = hyperparams['CRF_PARAMS']
    BiLSTM_STACK_CONFIG = hyperparams['BiLSTM_STACK_CONFIG']
    CLS2NER_KEYWORD_LEN = hyperparams['CLS2NER_KEYWORD_LEN']
    USE_ATTENTION_LAYER = hyperparams['USE_ATTENTION_LAYER']
    tag_size = hyperparams['ner_tag_lookuper'].size()
    label_size = hyperparams['cls_label_lookuper'].size()
    vocab_size = hyperparams['vocabulary_lookuper'].size()

    # input layer
    input_length = hyperparams['MAX_SENTENCE_LEN']
    input_layer = Input(shape=(input_length,), dtype='int32', name='input')

    # encoder
    if model_choice == 'VIRTUAL_EMBEDDING':  # cls_out embedding merged to ner_input_embedding as virtual embedding
        from mtnlpmodel.utils.model_util import VirtualEmbedding, Discriminator_new
        with tf.keras.backend.name_scope("Encoder"):
            embedding_layer_vocab = Embedding(vocab_size,
                                              EMBED_DIM,
                                              mask_zero=True,
                                              input_length=input_length,
                                              name='embedding_vocab'
                                              )
            embedding_layer_virtual = VirtualEmbedding(label_size,
                                                       EMBED_DIM,
                                                       mask_zero=True,
                                                       input_length=CLS2NER_KEYWORD_LEN,
                                                       mask_length=CLS2NER_KEYWORD_LEN,
                                                       name='embedding_virtual',
                                                       )

            embedding = embedding_layer_vocab(input_layer)

            embedding = Dropout(0.15)(embedding)  # just like random erase

        with tf.keras.backend.name_scope("Feature_extractor"):
            for bilstm_config in BiLSTM_STACK_CONFIG:
                biLSTM = Bidirectional(LSTM(return_sequences=True, **bilstm_config, name='biLSTM'))
            bilstm_extrator = biLSTM

        # classification branch
        ner_cls_layer, cls_output, cls_vector, _ = cls_branch_single_input(hyperparams['Arcloss'],
                                                           label_size, bilstm_extrator,
                                                           embedding,
                                                           outputlayer_name='cls_')
        ner_cls_output_shape = get_ner_cls_output_tensor_merge_embedding(CLS2NER_KEYWORD_LEN)(ner_cls_layer).shape
        ner_cls_output_layer = Lambda(get_ner_cls_output_tensor_merge_embedding(CLS2NER_KEYWORD_LEN),
                                      ner_cls_output_shape)(ner_cls_layer)

        # classification output will be used as a keyword adding to input of NER
        discriminator = Discriminator_new(onetask_output_shape=(CLS2NER_KEYWORD_LEN,),
                                          output_dtype='int32')
        ner_cls_input_layer = discriminator(ner_cls_output_layer)
        ner_virtual_embedding = embedding_layer_virtual(ner_cls_input_layer)
        ner_merged_embedding = tf.keras.layers.concatenate([ner_virtual_embedding, embedding], axis=1)
        ner_branch_embedding = ner_merged_embedding

        ner_feature_layer = None

    elif model_choice=='CLS2NER_INPUT':   # cls_out merged to ner_input as virtual keywords
        from mtnlpmodel.utils.model_util import Discriminator
        from mtnlpmodel.utils.input_process_util import build_vacablookuper_from_list
        vocabs = list(hyperparams['vocabulary_lookuper'].inverse_index_table.values())
        cls_labels = list(hyperparams['cls_label_lookuper'].inverse_index_table.values())
        vocabs.extend(cls_labels)
        vocabulary_lookuper = build_vacablookuper_from_list(*vocabs)

        vocab_size = vocabulary_lookuper.size()
        with tf.keras.backend.name_scope("Encoder"):
            embedding_layer = Embedding(vocab_size,
                                        EMBED_DIM,
                                        mask_zero=True,
                                        input_length=input_length,
                                        )
            embedding = embedding_layer(input_layer)

            embedding = Dropout(0.15)(embedding)  # just like random erase

        with tf.keras.backend.name_scope("Feature_extractor"):
            for bilstm_config in BiLSTM_STACK_CONFIG:
                biLSTM = Bidirectional(LSTM(return_sequences=True, **bilstm_config, name='biLSTM'))
            bilstm_extrator = biLSTM

        # classification branch
        ner_cls_layer, cls_output, cls_vector, _ = cls_branch_single_input(hyperparams['Arcloss'],
                                                                       label_size, bilstm_extrator, embedding,
                                                                       outputlayer_name='cls_')
        ner_cls_output_shape = get_ner_cls_output_tensor_merge_input(CLS2NER_KEYWORD_LEN,
                                                                     **{"vocab_size":vocab_size,
                                                                        "label_size":label_size})(ner_cls_layer).shape

        ner_cls_output_layer = Lambda(get_ner_cls_output_tensor_merge_input(
                                         CLS2NER_KEYWORD_LEN,
                                         **{"vocab_size": vocab_size, "label_size": label_size}),
                                      ner_cls_output_shape)(ner_cls_layer)

        # classification output will be used as a keyword adding to input of NER
        discriminator = Discriminator(input_layer, onetask_output_shape=(CLS2NER_KEYWORD_LEN,),
                                      output_dtype='int32')
        merged_ner_input_layer = discriminator([ner_cls_output_layer, input_layer])
        ner_branch_embedding = embedding_layer(merged_ner_input_layer)

        ner_feature_layer = None

    else: # task independent
        with tf.keras.backend.name_scope("Encoder"):
            embedding_layer = Embedding(vocab_size,
                                        EMBED_DIM,
                                        mask_zero=True,
                                        input_length=input_length,
                                        )

            embedding = embedding_layer(input_layer)

            embedding = Dropout(0.15)(embedding)    # just like random erase

        with tf.keras.backend.name_scope("Feature_extractor"):
            for bilstm_config in BiLSTM_STACK_CONFIG:
                biLSTM = Bidirectional(LSTM(return_sequences=True, **bilstm_config, name='biLSTM'))
            bilstm_extrator = biLSTM

        # classification branch
        _, cls_output, cls_vector, cls_feature_layer = cls_branch_single_input(hyperparams['Arcloss'],
                                                            label_size, bilstm_extrator,
                                                            embedding, outputlayer_name='cls_')
        ner_branch_embedding = embedding
        ner_feature_layer = cls_feature_layer

    # NER branch
    with tf.keras.backend.name_scope("NER_branch"):
        if ner_feature_layer is None:
            # print_op = tf.print(ner_virtual_embedding._keras_mask, ner_embedding._keras_mask)
            # with tf.control_dependencies([print_op]):
            embedding_layer = LayerNormalization()(ner_branch_embedding)
            biLSTM = bilstm_extrator(embedding_layer)
            biLSTM = LayerNormalization()(biLSTM)
            if USE_ATTENTION_LAYER:
                biLSTM = GlobalAttentionLayer()(biLSTM)
            ner_output = CRF(tag_size, name="crf_", **CRF_PARAMS)(biLSTM)
        else:
            biLSTM = ner_feature_layer
            biLSTM = LayerNormalization()(biLSTM)
            if USE_ATTENTION_LAYER:
                biLSTM = GlobalAttentionLayer()(biLSTM)
            ner_output = CRF(tag_size, name="crf_", **CRF_PARAMS)(biLSTM)

    # merge NER and Classification
    model = Model(inputs=[input_layer], outputs=[ner_output, cls_output])
    semantic_vector = Model(inputs=[input_layer], outputs=cls_vector)

    freeze_layers = [layer for layer in model.layers if layer.name in freeze_list]
    trainable_layers = [layer for layer in model.layers if layer.name not in freeze_list]
    
    for layer in freeze_layers:
        layer.trainable = False

    for layer in trainable_layers:
        layer.trainable = True
        
    model.load_weights(model_weights_path, by_name=True)
    
    return model, semantic_vector


def get_freeze_list_for_finetuning(model_choice):
    '''Different structure models have different layers and layer names,
       through this func to get the corresponding recommendation frozen list.
       Layers in frozen_list are not trainable during the finetuning process. 
       You can modify the return list to customize your own frozen list.
    '''
    if model_choice=='VIRTUAL_EMBEDDING':
        return ['bidirectional', 'embedding_vocab']
    elif model_choice=='CLS2NER_INPUT':
        return ['bidirectional']
    else:
        return ['embedding', 'bidirectional']
    

def finetuning_logger(*args):
    print('Fine-tuning processing: ')
    for arg in args:
        try:
            if os.path.split(arg)[-1].startswith('weights'):
                print('Load model weights from {}'.format(arg))
        except:
            if isinstance(arg, list):
                print('Frozen list is [{}]'.format(', '.join(arg)))


import tensorflow as tf
from tensorflow import keras
from tensorflow.python.keras.callbacks import Callback
import tensorflow.keras.backend as K


class SetLearningRate:
    """
        层的一个包装，用来设置当前层的学习率

        转载自：https://kexue.fm/archives/6418
    """

    def __init__(self, layer, lr=0.01, is_ada=False):
        self.layer = layer
        self.lamb = lr  # 学习率比例
        self.is_ada = is_ada  # 是否自适应学习率优化器

    def __call__(self, inputs):
        with tf.keras.backend.name_scope(self.layer.name):
            if not self.layer.built:
                input_shape = tf.keras.backend.int_shape(inputs)
                self.layer.build(input_shape)
                self.layer.built = True
                if self.layer._initial_weights is not None:
                    self.layer.set_weights(self.layer._initial_weights)
        for key in ['kernel', 'bias', 'embeddings', 'depthwise_kernel', 'pointwise_kernel', 'recurrent_kernel', 'gamma',
                    'beta']:
            if hasattr(self.layer, key):
                weight = getattr(self.layer, key)
                if self.is_ada:
                    lamb = self.lamb  # 自适应学习率优化器直接保持lamb比例
                else:
                    lamb = self.lamb ** 0.5  # SGD（包括动量加速），lamb要开平方
                tf.keras.backend.set_value(weight, tf.keras.backend.eval(weight) / lamb)  # 更改初始化
                # setattr(self.layer, key, weight*lamb)
                tf.keras.backend.set_value(weight, tf.keras.backend.eval(weight) * lamb)  # 修正

        return self.layer(inputs)


class LearningRateDecay(Callback):
    '''lr decay during training epoches
    '''

    def __init__(self, model):
        super().__init__()
        self.model = model

    def scheduler(self, epoch):
        # 每隔5个epoch，学习率减小为原来的1/10
        if epoch % 6 == 0 and epoch != 0:
            lr = K.get_value(self.model.optimizer.lr)
            # K.set_value(self.model.optimizer.lr, lr * 0.1)
            K.set_value(self.model.optimizer.lr, lr * 5)
            # print("Learning Rate changed to {}".format(lr * 0.1))
            print("Learning Rate changed to {}".format(lr * 5))
        return K.get_value(self.model.optimizer.lr)

    # if __name__ == '__main__':
    # how to use
    # from tensorflow.keras.callbacks import LearningRateScheduler
    # from mtnlp_model.mtnlpmodel.trainer.lrset_util import LearningRateDecay
    # reduce_lr = LearningRateScheduler(LearningRateDecay(model).scheduler)
    # callbacks_list.append(reduce_lr)




import numpy as np

class CyclicLR(Callback):
    '''
        https://github.com/bckenstler/CLR/blob/master/clr_callback.py
    '''

    """This callback implements a cyclical learning rate policy (CLR).
    The method cycles the learning rate between two boundaries with
    some constant frequency, as detailed in this paper (https://arxiv.org/abs/1506.01186).
    The amplitude of the cycle can be scaled on a per-iteration or
    per-cycle basis.
    This class has three built-in policies, as put forth in the paper.
    "triangular":
        A basic triangular cycle w/ no amplitude scaling.
    "triangular2":
        A basic triangular cycle that scales initial amplitude by half each cycle.
    "exp_range":
        A cycle that scales initial amplitude by gamma**(cycle iterations) at each
        cycle iteration.
    For more detail, please see paper.

    # Example
        ```python
            clr = CyclicLR(base_lr=0.001, max_lr=0.006,
                                step_size=2000., mode='triangular')
            model.fit(X_train, Y_train, callbacks=[clr])
        ```

    Class also supports custom scaling functions:
        ```python
            clr_fn = lambda x: 0.5*(1+np.sin(x*np.pi/2.))
            clr = CyclicLR(base_lr=0.001, max_lr=0.006,
                                step_size=2000., scale_fn=clr_fn,
                                scale_mode='cycle')
            model.fit(X_train, Y_train, callbacks=[clr])
        ```
    # Arguments
        base_lr: initial learning rate which is the
            lower boundary in the cycle.
        max_lr: upper boundary in the cycle. Functionally,
            it defines the cycle amplitude (max_lr - base_lr).
            The lr at any cycle is the sum of base_lr
            and some scaling of the amplitude; therefore
            max_lr may not actually be reached depending on
            scaling function.
        step_size: number of training iterations per
            half cycle. Authors suggest setting step_size
            2-8 x training iterations in epoch.
        mode: one of {triangular, triangular2, exp_range}.
            Default 'triangular'.
            Values correspond to policies detailed above.
            If scale_fn is not None, this argument is ignored.
        gamma: constant in 'exp_range' scaling function:
            gamma**(cycle iterations)
        scale_fn: Custom scaling policy defined by a single
            argument lambda function, where
            0 <= scale_fn(x) <= 1 for all x >= 0.
            mode paramater is ignored
        scale_mode: {'cycle', 'iterations'}.
            Defines whether scale_fn is evaluated on
            cycle number or cycle iterations (training
            iterations since start of cycle). Default is 'cycle'.
    """

    def __init__(self, base_lr=0.001, max_lr=0.006, step_size=2000., mode='triangular',
                 gamma=1., scale_fn=None, scale_mode='cycle'):
        super(CyclicLR, self).__init__()

        self.base_lr = base_lr
        self.max_lr = max_lr
        self.step_size = step_size
        self.mode = mode
        self.gamma = gamma
        if scale_fn == None:
            if self.mode == 'triangular':
                self.scale_fn = lambda x: 1.
                self.scale_mode = 'cycle'
            elif self.mode == 'triangular2':
                self.scale_fn = lambda x: 1 / (2. ** (x - 1))
                self.scale_mode = 'cycle'
            elif self.mode == 'exp_range':
                self.scale_fn = lambda x: gamma ** (x)
                self.scale_mode = 'iterations'
        else:
            self.scale_fn = scale_fn
            self.scale_mode = scale_mode
        self.clr_iterations = 0.
        self.trn_iterations = 0.
        self.history = {}

        self._reset()

    def _reset(self, new_base_lr=None, new_max_lr=None,
               new_step_size=None):
        """Resets cycle iterations.
        Optional boundary/step size adjustment.
        """
        if new_base_lr != None:
            self.base_lr = new_base_lr
        if new_max_lr != None:
            self.max_lr = new_max_lr
        if new_step_size != None:
            self.step_size = new_step_size
        self.clr_iterations = 0.

    def clr(self):
        cycle = np.floor(1 + self.clr_iterations / (2 * self.step_size))
        x = np.abs(self.clr_iterations / self.step_size - 2 * cycle + 1)
        if self.scale_mode == 'cycle':
            return self.base_lr + (self.max_lr - self.base_lr) * np.maximum(0, (1 - x)) * self.scale_fn(cycle)
        else:
            return self.base_lr + (self.max_lr - self.base_lr) * np.maximum(0, (1 - x)) * self.scale_fn(
                self.clr_iterations)

    def on_train_begin(self, logs={}):
        logs = logs or {}

        if self.clr_iterations == 0:
            K.set_value(self.model.optimizer.lr, self.base_lr)
        else:
            K.set_value(self.model.optimizer.lr, self.clr())

    def on_batch_end(self, epoch, logs=None):

        logs = logs or {}
        self.trn_iterations += 1
        self.clr_iterations += 1

        self.history.setdefault('lr', []).append(K.get_value(self.model.optimizer.lr))
        self.history.setdefault('iterations', []).append(self.trn_iterations)

        for k, v in logs.items():
            self.history.setdefault(k, []).append(v)

        K.set_value(self.model.optimizer.lr, self.clr())


class SGDRScheduler(Callback):
    '''Schedule learning rates with restarts
     A simple restart technique for stochastic gradient descent.
    The learning rate decays after each batch and peridically resets to its
    initial value. Optionally, the learning rate is additionally reduced by a
    fixed factor at a predifined set of epochs.
     # Arguments
        epochsize: Number of samples per epoch during training.
        batchsize: Number of samples per batch during training.
        start_epoch: First epoch where decay is applied.
        epochs_to_restart: Initial number of epochs before restarts.
        mult_factor: Increase of epochs_to_restart after each restart.
        lr_fac: Decrease of learning rate at epochs given in
                lr_reduction_epochs.
        lr_reduction_epochs: Fixed list of epochs at which to reduce
                             learning rate.
     # References
        - [SGDR: Stochastic Gradient Descent with Restarts](http://arxiv.org/abs/1608.03983)
    '''

    def __init__(self,
                 epochsize,
                 batchsize,
                 epochs_to_restart=2,
                 mult_factor=2,
                 lr_fac=0.1,
                 lr_reduction_epochs=(60, 120, 160)):
        super(SGDRScheduler, self).__init__()
        self.epoch = -1
        self.batch_since_restart = 0
        self.next_restart = epochs_to_restart
        self.epochsize = epochsize
        self.batchsize = batchsize
        self.epochs_to_restart = epochs_to_restart
        self.mult_factor = mult_factor
        self.batches_per_epoch = self.epochsize / self.batchsize
        self.lr_fac = lr_fac
        self.lr_reduction_epochs = lr_reduction_epochs
        self.lr_log = []

    def on_train_begin(self, logs={}):
        self.lr = K.get_value(self.model.optimizer.lr)

    def on_epoch_begin(self, epoch, logs={}):
        self.epoch += 1

    def on_batch_end(self, batch, logs={}):
        fraction_to_restart = self.batch_since_restart / \
                              (self.batches_per_epoch * self.epochs_to_restart)
        lr = 0.5 * self.lr * (1 + np.cos(fraction_to_restart * np.pi))
        K.set_value(self.model.optimizer.lr, lr)

        self.batch_since_restart += 1
        self.lr_log.append(lr)

    def on_epoch_end(self, epoch, logs={}):
        if self.epoch + 1 == self.next_restart:
            self.batch_since_restart = 0
            self.epochs_to_restart *= self.mult_factor
            self.next_restart += self.epochs_to_restart

        if (self.epoch + 1) in self.lr_reduction_epochs:
            self.lr *= self.lr_fac



#-*- coding: UTF-8 -*

import warnings
warnings.filterwarnings('ignore')
import tqdm
import json, os, time
import tensorflow as tf
from deliverable_model import load
from deliverable_model.request import Request
from mtnlpmodel.utils.model_util import ArcFace
from mtnlpmodel.utils.model_util import VirtualEmbedding
from tokenizer_tools.tagset.offset.corpus import Corpus
from mtnlpmodel.utils.model_util import Mish
from tokenizer_tools.utils.batch_iterator import BatchingIterator

tf.keras.utils.get_custom_objects()[Mish.__name__] = Mish
tf.keras.utils.get_custom_objects()[ArcFace.__name__] = ArcFace
tf.keras.utils.get_custom_objects()[VirtualEmbedding.__name__] = VirtualEmbedding

'''
    调用模型对输入数据进行推理，即采用模型对输入数据做实体识别标注，
    推理结果供人工复查。
'''

class MtModelInference_Deliverable:
    def __init__(self, config_filepath):
        self.config_filepath = config_filepath
        self.config = self.get_config()

        if not os.path.exists(self.config['output_filepath']):
            os.mkdir(self.config['output_filepath'])


    def get_config(self):
        '''
            从配置文件中读取配置信息
        '''
        with open(self.config_filepath, 'rb') as f:
            self.config = json.load(f)
        return self.config


    @staticmethod
    def generate_batch_input(input_data, batch_size):
        length = len(input_data)
        if length<batch_size:
            yield input_data
        else:
            for i in range(0, length, batch_size):
                yield(input_data[i: i+batch_size])


    def _inference(self, model, input_data):
        output = []
        batch_size = 1
        batches = MtModelInference_Deliverable.generate_batch_input(input_data, batch_size)
        # batch_generator = BatchingIterator(batch_size)
        # batches = batch_generator(input_data)

        for batch in tqdm.tqdm(batches):
            request = Request(batch)
            response = model.inference(request)
            tmp_result = response['data'][0].sequence
            tmp_result.label = response['cls'][0][0]
            output.append(tmp_result)

        predict_result = Corpus(output)
        predict_result.write_to_file(os.path.join(self.config['output_filepath'], 'inference_out.conllx'))

        print('*** inference has been done, please check the result through the path below:')
        print('==> {}'.format(self.config['output_filepath']))

        return


    def call_inference(self):
        input_rawdata_filename = self.config['input_rawdata']
        input_file = os.path.join(self.config['data_filepath'], input_rawdata_filename)
        with open(input_file,'rt',encoding='utf-8') as f:
            input_data = [line.strip() for line in f.readlines()]
        model = load(self.config['model_filepath'])
        self._inference(model, input_data)


if __name__ == "__main__":
    print('*****The Inference program is running now, please wait.*****\n')
    start = time.time()
    config_filepath = './configure.json'
    MtModelInference_Deliverable(config_filepath).call_inference()
    end = time.time()
    print('==> Time cost is', end - start, 's')


import copy
from typing import List
from collections import namedtuple
from pathlib import Path
from tokenizer_tools.tagset.NER.base_tagset import BaseTagSet
from tokenizer_tools.tagset.exceptions import TagSetDecodeError
from tokenizer_tools.tagset.offset.document import Document
from tokenizer_tools.tagset.offset.span import Span
from tokenizer_tools.tagset.offset.span_set import SpanSet
from deliverable_model.processor_base import ProcessorBase
from deliverable_model.request import Request
from deliverable_model.response import Response



def tags_to_span_set(tags: List[str]) -> SpanSet:
    decoder = BIOSequenceEncoderDecoder()
    span_info = decoder.decode_to_offset(tags)

    return SpanSet([Span(i[0], i[1], i[2]) for i in span_info])


class BIOEncoderDecoder(BaseTagSet):
    """
    Encoder and Decoder for BILUO scheme
    """

    # O is very easy confused with zero, using oscar instead in the code
    oscar = "O"

    def generate_tag(self, prefix):
        if self.tag_name == self.oscar:
            # O tag is very special, it always return O
            return self.oscar

        if self.tag_name is not None:
            tag = "{}-{}".format(prefix, self.tag_name)
        else:
            # if tag_name is None, no more tag_name in tag
            tag = prefix

        return tag

    def encode(self, sequence):
        len_of_sequence = len(sequence)

        if len_of_sequence == 1:
            return [self.generate_tag("B")]

        elif len_of_sequence == 2:
            return [self.generate_tag("B"), self.generate_tag("I")]

        else:
            return (
                [self.generate_tag("B")]
                + [self.generate_tag("I")] * (len_of_sequence - 2)
                + [self.generate_tag("I")]
            )

    def decode(self, sequence):
        pass

    def all_tag_set(self):
        tag_set = {self.generate_tag(i) for i in "BI"}
        tag_set_oscar = {self.oscar}
        tag_set.update(tag_set_oscar)
        return tag_set


class BIOSequenceEncoderDecoder(object):
    # O is very easy confused with zero, using oscar instead in the code
    oscar = "O"

    legal_set = {
        ("B", "I"),
        ("I", "I"),
        (oscar, "B"),
        ("I", oscar),
        ("B", oscar),
        ("I", "B"),
        ("B", "B"),
    }

    prefix_set = set("BI")

    def __init__(self, *args, **kwargs):
        self.ignore_error = kwargs.get("ignore_error", True)

    def parse_tag(self, tag):
        # TODO: already replaced by inline code, remove me in later version
        if tag == self.oscar:
            return self.oscar, None

        # set maxsplit to 1, so raw_tag_name can contains '-' char legally
        raw_prefix, raw_tag_name = tag.split("-", maxsplit=1)

        prefix = raw_prefix.strip()
        tag_name = raw_tag_name.strip()

        if prefix and tag_name and prefix in self.prefix_set:
            return prefix, tag_name

        raise ValueError("tag: {} is not a avoid tag".format(tag))

    def is_prefix_legal(self, previous, current):
        node = (previous, current)

        return node in self.legal_set

    def decode_to_offset(self, sequence):
        offset_list = []

        last_tag_prefix = None
        next_tag_prefix = None
        tag_length = 0
        tag_name_cache = None

        for index, item in enumerate(sequence):
            # inline function <<< self.parse_tag(item)
            tag = item
            if tag == self.oscar:
                prefix = self.oscar
                tag_name = None
            else:
                # set maxsplit to 1, so raw_tag_name can contains '-' char legally
                prefix, tag_name = tag.split("-", maxsplit=1)

                if prefix and tag_name and prefix in self.prefix_set:
                    pass
                else:
                    raise ValueError("tag: {} is not a avoid tag".format(tag))
            # inline function >>> self.parse_tag(item)

            if last_tag_prefix is None:
                if prefix == self.oscar:
                    # ignore it
                    continue
                elif prefix == "B":
                    if index<=len(sequence)-1:
                        if len(sequence[index+1])>1:
                            next_tag_prefix,_ = sequence[index+1].split("-", maxsplit=1)
                        else:
                            next_tag_prefix = sequence[index+1]
                    else:
                        next_tag_prefix = None
                    if (next_tag_prefix is None
                        or next_tag_prefix=='O'
                        or next_tag_prefix=='B'):
                        offset_list.append((index, index + 1, tag_name))
                    else:
                        tag_name_cache = tag_name
                        last_tag_prefix = prefix
                        tag_length += 1
                else:
                    if not self.ignore_error:
                        raise TagSetDecodeError(
                            "sequence: {} is not a valid tag sequence".format(
                                sequence[: index + 1]
                            )
                        )
                    else:
                        continue
            else:
                if not self.is_prefix_legal(last_tag_prefix, prefix):
                    raise TagSetDecodeError(
                        "sequence: {} is not a valid tag sequence".format(
                            sequence[: index + 1]
                        )
                    )

                if prefix == "I":
                    if index<=len(sequence)-1:
                        if len(sequence[index+1])>1:
                            next_tag_prefix,_ = sequence[index+1].split("-", maxsplit=1)
                        else:
                            next_tag_prefix = sequence[index+1]
                    else:
                        next_tag_prefix = None

                    if (next_tag_prefix is None
                        or next_tag_prefix=='O'
                        or next_tag_prefix=='B'):
                        if tag_name_cache == tag_name:
                            offset_list.append(
                                (index - tag_length, index + 1, tag_name_cache)
                            )

                            # clean up
                            last_tag_prefix = None
                            tag_length = 0
                            tag_name_cache = None
                            next_tag_prefix = None
                        else:
                            raise TagSetDecodeError(
                                "sequence: {} is not a valid tag sequence".format(
                                    sequence[: index + 1]
                                )
                            )
                    elif prefix == "I":
                        if tag_name_cache == tag_name:
                            last_tag_prefix = prefix
                            tag_length += 1
                        else:
                            raise TagSetDecodeError(
                                "sequence: {} is not a valid tag sequence".format(
                                    sequence[: index + 1]
                                )
                            )

        return offset_list

    def to_offset(self, sequence, text, **kwargs):
        span_set = SpanSet(
            [
                Span(offset[0], offset[1], offset[2])
                for offset in self.decode_to_offset(sequence)
            ]
        )

        label = kwargs.pop("label", None)
        id_ = kwargs.pop("id", None)
        extra_attr = kwargs

        seq = Document(text, span_set, id_, label, extra_attr)
        seq.span_set.bind(seq)

        return seq


PredictResult = namedtuple("PredictResult", ["sequence", "is_failed", "exec_msg"])

class BIOEncodeProcessor(ProcessorBase):
    def __init__(self, decoder=None, **kwargs):
        super().__init__(**kwargs)

        self.decoder = decoder
        self.request_query = None

    @classmethod
    def load(cls, parameter: dict, asset_dir) -> "ProcessorBase":

        decoder = BIOSequenceEncoderDecoder()

        self = cls(decoder, **parameter)

        return self

    def preprocess(self, request: Request) -> Request:
        # record request for postprocess usage
        self.request_query = copy.deepcopy(request[self.pre_input_key])

        # don't change the request
        return request

    def postprocess(self, response: Response) -> Response:
        from tokenizer_tools.tagset.exceptions import TagSetDecodeError
        from tokenizer_tools.tagset.offset.sequence import Sequence

        tags_list = response[self.post_input_key]
        raw_text_list = self.request_query

        infer_result = []

        for raw_text, tags in zip(raw_text_list, tags_list):
            # decode Unicode
            tags_seq = [i.decode() if isinstance(i, bytes) else i for i in tags]

            # BILUO to offset
            is_failed = False
            exec_msg = None
            try:
                seq = self.decoder.to_offset(tags_seq, raw_text)
            except TagSetDecodeError as e:
                exec_msg = str(e)

                # invalid tag sequence will raise exception
                # so return a empty result to avoid batch fail
                seq = Sequence(raw_text)
                is_failed = True

            infer_result.append(PredictResult(seq, is_failed, exec_msg))

        response[self.post_output_key] = infer_result

        return response

    def get_dependency(self) -> list:
        return ["tokenizer_tools"]


if __name__ == "__main__":
    decoder = BIOSequenceEncoderDecoder()
    result = decoder.decode_to_offset(["U-XX"])
    print(result)
    assert result == [(0, 1, "XX")]

    result = decoder.decode_to_offset(["U-XX", "U-YY"])
    print(result)
    assert result == [(0, 1, "XX"), (1, 2, "YY")]

    result = decoder.decode_to_offset(["B-XX", "I-XX", "L-XX"])
    print(result)
    assert result == [(0, 3, "XX")]
